import freeOrionAIInterface as fo
from logging import debug, error, info, warning
from operator import itemgetter

import AIDependencies
import AIstate
import ExplorationAI
import FleetUtilsAI
import PlanetUtilsAI
from AIDependencies import INVALID_ID, OUTPOSTING_TECH
from aistate_interface import get_aistate
from colonization.calculate_planet_colonization_rating import (
    calculate_planet_colonization_rating,
)
from colonization.colony_score import MINIMUM_COLONY_SCORE
from colonization.planet_supply import get_planet_supply
from common.print_utils import Number, Sequence, Table, Text
from empire.colony_builders import can_build_colony_for_species, get_colony_builders
from EnumsAI import EmpireProductionTypes, MissionType, PriorityType, ShipRoleType
from expansion_plans import (
    get_colonisable_outpost_ids,
    get_colonisable_planet_ids,
    update_colonisable_outpost_ids,
    update_colonisable_planet_ids,
)
from freeorion_tools import get_ship_part
from freeorion_tools.caching import cache_for_current_turn
from freeorion_tools.timers import AITimer
from target import TargetPlanet
from turn_state import (
    get_colonized_planets_in_system,
    get_empire_outposts,
    get_number_of_colonies,
    get_unowned_empty_planets,
)
from turn_state.design import get_best_ship_info

colonization_timer = AITimer("getColonyFleets()")

_all_colony_opportunities = {}


@cache_for_current_turn
def colony_pod_cost_turns():
    empire = fo.getEmpire()
    empire_id = empire.empireID
    loc = INVALID_ID
    pid = INVALID_ID
    parts = [get_ship_part(part) for part in list(empire.availableShipParts)]
    colo_parts = [
        part for part in parts if part.partClass in frozenset({fo.shipPartClass.colony}) and part.capacity > 0
    ]
    if colo_parts:
        colo_part = max(colo_parts, key=lambda x: x.capacity)
        base_cost = colo_part.productionCost(empire_id, pid, loc)
        build_turns = colo_part.productionTime(empire_id, pid, loc)
    else:
        base_cost = 0
        build_turns = 0
        debug("no available colony parts with capacity > 0")
    return (base_cost * (1 + get_number_of_colonies() * AIDependencies.COLONY_POD_UPKEEP), build_turns)


def outpod_pod_cost():
    return AIDependencies.OUTPOST_POD_COST * (1 + get_number_of_colonies() * AIDependencies.COLONY_POD_UPKEEP)


def galaxy_is_sparse():
    setup_data = fo.getGalaxySetupData()
    avg_empire_systems = setup_data.size // len(fo.allEmpireIDs())
    return (setup_data.monsterFrequency <= fo.galaxySetupOptionMonsterFreq.veryLow) and (
        (avg_empire_systems >= 40) or ((avg_empire_systems >= 35) and (setup_data.shape != fo.galaxyShape.elliptical))
    )


def get_colony_fleets():
    """examines known planets, collects various colonization data, to be later used to send colony fleets"""
    universe = fo.getUniverse()
    empire = fo.getEmpire()

    colonization_timer.start("Identify Existing colony/outpost targets")
    colony_targeted_planet_ids = FleetUtilsAI.get_targeted_planet_ids(universe.planetIDs, MissionType.COLONISATION)
    all_colony_targeted_system_ids = PlanetUtilsAI.get_systems(colony_targeted_planet_ids)
    colony_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.COLONISATION)
    num_colony_fleets = len(FleetUtilsAI.extract_fleet_ids_without_mission_types(colony_fleet_ids))

    outpost_targeted_planet_ids = FleetUtilsAI.get_targeted_planet_ids(universe.planetIDs, MissionType.OUTPOST)
    outpost_targeted_planet_ids.extend(
        FleetUtilsAI.get_targeted_planet_ids(universe.planetIDs, MissionType.ORBITAL_OUTPOST)
    )
    all_outpost_targeted_system_ids = PlanetUtilsAI.get_systems(outpost_targeted_planet_ids)
    outpost_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.OUTPOST)
    num_outpost_fleets = len(FleetUtilsAI.extract_fleet_ids_without_mission_types(outpost_fleet_ids))

    debug("Colony Targeted SystemIDs: %s" % all_colony_targeted_system_ids)
    debug("Colony Targeted PlanetIDs: %s" % colony_targeted_planet_ids)
    debug(colony_fleet_ids and "Colony Fleet IDs: %s" % colony_fleet_ids or "Available Colony Fleets: 0")
    debug("Colony Fleets Without Missions: %s" % num_colony_fleets)
    debug("")
    debug("Outpost Targeted SystemIDs: %s" % all_outpost_targeted_system_ids)
    debug("Outpost Targeted PlanetIDs: %s" % outpost_targeted_planet_ids)
    debug(outpost_fleet_ids and "Outpost Fleet IDs: %s" % outpost_fleet_ids or "Available Outpost Fleets: 0")
    debug("Outpost Fleets Without Missions: %s" % num_outpost_fleets)
    debug("")

    # export targeted systems for other AI modules
    AIstate.colonyTargetedSystemIDs = all_colony_targeted_system_ids
    AIstate.outpostTargetedSystemIDs = all_outpost_targeted_system_ids

    colonization_timer.start("Identify colony base targets")
    # keys are sets of ints; data is doubles
    available_pp = {tuple(el.key()): el.data() for el in empire.planetsWithAvailablePP}

    avail_pp_by_sys = {}
    for p_set in available_pp:
        avail_pp_by_sys.update([(sys_id, available_pp[p_set]) for sys_id in set(PlanetUtilsAI.get_systems(p_set))])

    evaluated_colony_planet_ids = list(
        get_unowned_empty_planets().union(get_empire_outposts()) - set(colony_targeted_planet_ids)
    )  # places for possible colonyBase

    aistate = get_aistate()
    outpost_base_manager = aistate.orbital_colonization_manager

    for pid in evaluated_colony_planet_ids:  # TODO: reorganize
        planet = universe.getPlanet(pid)
        if not planet:
            continue
        sys_id = planet.systemID
        for pid2 in get_colonized_planets_in_system(sys_id):
            planet2 = universe.getPlanet(pid2)
            if not (planet2 and can_build_colony_for_species(planet2.speciesName)):
                continue
            if planet.unowned:
                outpost_base_manager.create_new_plan(pid, pid2)

    colonization_timer.start("Initiate outpost base construction")

    reserved_outpost_base_targets = outpost_base_manager.get_targets()
    debug("Current qualifyingOutpostBaseTargets: %s" % reserved_outpost_base_targets)
    outpost_base_manager.build_bases()

    colonization_timer.start("Evaluate All Colony Opportunities")
    evaluated_outpost_planet_ids = list(
        get_unowned_empty_planets()
        - set(outpost_targeted_planet_ids)
        - set(colony_targeted_planet_ids)
        - set(reserved_outpost_base_targets)
    )
    _all_colony_opportunities.clear()
    _all_colony_opportunities.update(
        assign_colonisation_values(evaluated_colony_planet_ids, MissionType.COLONISATION, None, return_all=True)
    )
    evaluated_colony_planets = {pid: values[0] for pid, values in _all_colony_opportunities.items()}
    colonization_timer.start("Evaluate Outpost Opportunities")

    sorted_planets = list(evaluated_colony_planets.items())
    sorted_planets.sort(key=itemgetter(1), reverse=True)

    _print_colony_candidate_table(sorted_planets, show_detail=False)

    sorted_planets = [(planet_id, score[:2]) for planet_id, score in sorted_planets if score[0] > 0]
    # export planets for other AI modules
    update_colonisable_planet_ids(sorted_planets)

    evaluated_outpost_planets = assign_colonisation_values(evaluated_outpost_planet_ids, MissionType.OUTPOST, None)
    # if outposted planet would be in supply range, let outpost value be best of outpost value or colonization value
    for pid in set(evaluated_outpost_planets).intersection(evaluated_colony_planets):
        if get_planet_supply(pid, -99) >= 0:
            evaluated_outpost_planets[pid] = (
                max(evaluated_colony_planets[pid][0], evaluated_outpost_planets[pid][0]),
                "",
            )

    colonization_timer.stop()

    sorted_outposts = list(evaluated_outpost_planets.items())
    sorted_outposts.sort(key=itemgetter(1), reverse=True)

    _print_outpost_candidate_table(sorted_outposts)

    sorted_outposts = [(planet_id, score[:2]) for planet_id, score in sorted_outposts if score[0] > 0]
    # export outposts for other AI modules
    update_colonisable_outpost_ids(sorted_outposts)
    colonization_timer.stop_print_and_clear()


# TODO: clean up suppliable versus annexable
def assign_colonisation_values(planet_ids, mission_type, species, detail=None, return_all=False):
    """Creates a dictionary that takes planetIDs as key and their colonisation score as value."""
    if detail is None:
        detail = []
    orig_detail = detail
    planet_values = {}
    if mission_type == MissionType.OUTPOST:
        # print "\n=========\nAssigning Outpost Values\n========="
        try_species = [""]
    elif species is not None:
        # print "\n=========\nAssigning Colony Values\n========="
        if isinstance(species, str):
            try_species = [species]
        elif isinstance(species, list):
            try_species = species
        else:
            try_species = [species.name]
    else:
        # print "\n=========\nAssigning Colony Values\n========="
        try_species = list(get_colony_builders())
    for planet_id in planet_ids:
        pv = []
        for spec_name in try_species:
            # TODO: this function never returned detail to its caller, can we remove it?
            detail = orig_detail[:]
            # appends (score, species_name, detail)
            pv.append(
                (
                    calculate_planet_colonization_rating(
                        planet_id=planet_id,
                        mission_type=mission_type,
                        spec_name=spec_name,
                        detail=detail,
                    ),
                    spec_name,
                    detail,
                )
            )
        all_sorted = sorted(pv, reverse=True)
        best = all_sorted[:1]
        if best:
            if return_all:
                planet_values[planet_id] = all_sorted
            else:
                planet_values[planet_id] = best[0]
                # print best[0][2]
    return planet_values


def assign_colony_fleets_to_colonise():
    aistate = get_aistate()
    aistate.orbital_colonization_manager.assign_bases_to_colonize()

    # assign fleet targets to colonisable planets
    all_colony_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.COLONISATION)
    send_colony_ships(
        FleetUtilsAI.extract_fleet_ids_without_mission_types(all_colony_fleet_ids),
        list(get_colonisable_planet_ids().items()),
        MissionType.COLONISATION,
    )

    # assign fleet targets to colonisable outposts
    all_outpost_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.OUTPOST)
    send_colony_ships(
        FleetUtilsAI.extract_fleet_ids_without_mission_types(all_outpost_fleet_ids),
        list(get_colonisable_outpost_ids().items()),
        MissionType.OUTPOST,
    )


def send_colony_ships(colony_fleet_ids, evaluated_planets, mission_type):  # noqa: C901
    """sends a list of colony ships to a list of planet_value_pairs"""
    fleet_pool = colony_fleet_ids[:]
    try_all = False
    if mission_type == MissionType.OUTPOST:
        cost = 20 + outpod_pod_cost()
    else:
        try_all = True
        cost = 20 + colony_pod_cost_turns()[1]
        if fo.currentTurn() < 50:
            cost *= 0.4  # will be making fast tech progress so value is underestimated
        elif fo.currentTurn() < 80:
            cost *= 0.8  # will be making fast-ish tech progress so value is underestimated

    potential_targets = [
        (pid, (score, specName))
        for (pid, (score, specName)) in evaluated_planets
        if score > (0.8 * cost) and score > MINIMUM_COLONY_SCORE
    ]

    debug(f"Colony/outpost ship matching: fleets {fleet_pool} to planets {evaluated_planets}")

    if try_all:
        debug("Trying best matches to current colony ships")
        best_scores = dict(evaluated_planets)
        potential_targets = []
        for pid, ratings in _all_colony_opportunities.items():
            for rating in ratings:
                if rating[0] >= 0.75 * best_scores.get(pid, [9999])[0]:
                    potential_targets.append((pid, rating))
        potential_targets.sort(key=itemgetter(1), reverse=True)

    # added a lot of checking because have been getting mysterious exception, after too many recursions to get info
    fleet_pool = set(fleet_pool)
    universe = fo.getUniverse()
    for fid in fleet_pool:
        fleet = universe.getFleet(fid)
        if not fleet or fleet.empty:
            warning("Bad fleet ( ID %d ) given to colonization routine; will be skipped" % fid)
            fleet_pool.remove(fid)
            continue
        report_str = "Fleet ID (%d): %d ships; species: " % (fid, fleet.numShips)
        for sid in fleet.shipIDs:
            ship = universe.getShip(sid)
            if not ship:
                report_str += "NoShip, "
            else:
                report_str += "%s, " % ship.speciesName
        debug(report_str)
    debug("")
    already_targeted = []
    # for planetID_value_pair in evaluatedPlanets:
    aistate = get_aistate()
    while fleet_pool and potential_targets:
        target = potential_targets.pop(0)
        if target in already_targeted:
            continue
        planet_id = target[0]
        if planet_id in already_targeted:
            continue
        planet = universe.getPlanet(planet_id)
        sys_id = planet.systemID
        if (
            aistate.systemStatus.setdefault(sys_id, {}).setdefault("monsterThreat", 0) > 2000
            or fo.currentTurn() < 20
            and aistate.systemStatus[sys_id]["monsterThreat"] > 200
        ):
            debug(
                "Skipping colonization of system %s due to Big Monster, threat %d"
                % (universe.getSystem(sys_id), aistate.systemStatus[sys_id]["monsterThreat"])
            )
            already_targeted.append(planet_id)
            continue

        # make sure not to run into stationary guards
        if ExplorationAI.system_could_have_unknown_stationary_guard(sys_id):
            ExplorationAI.request_emergency_exploration(sys_id)
            continue

        this_spec = target[1][1]
        found_fleets = []
        try:
            this_fleet_list = FleetUtilsAI.get_fleets_for_mission(
                target_stats={},
                min_stats={},
                cur_stats={},
                starting_system=sys_id,
                species=this_spec,
                fleet_pool_set=fleet_pool,
                fleet_list=found_fleets,
            )
        except Exception as e:
            error(e, exc_info=True)
            continue
        if not this_fleet_list:
            fleet_pool.update(found_fleets)  # just to be safe
            continue  # must have no compatible colony/outpost ships
        fleet_id = this_fleet_list[0]
        already_targeted.append(planet_id)
        ai_target = TargetPlanet(planet_id)
        aistate.get_fleet_mission(fleet_id).set_target(mission_type, ai_target)


def _print_outpost_candidate_table(candidates, show_detail=False):
    """Print a summary for the outpost candidates in a table format to log.

    :param candidates: list of (planet_id, (score, species, details)) tuples
    """
    __print_candidate_table(candidates, mission="Outposts", show_detail=show_detail)


def _print_colony_candidate_table(candidates, show_detail=False):
    """Print a summary for the colony candidates in a table format to log.

    :param candidates: list of (planet_id, (score, species, details)) tuples
    """
    __print_candidate_table(candidates, mission="Colonization", show_detail=show_detail)


def __print_candidate_table(candidates, mission, show_detail=False):
    universe = fo.getUniverse()
    if mission == "Colonization":
        first_column = Text("Score/Species")

        def get_first_column_value(x):
            return round(x[0], 2), x[1]

    elif mission == "Outposts":
        first_column = Number("Score")
        get_first_column_value = itemgetter(0)
    else:
        warning(f"__print_candidate_table({candidates}, {mission}): Invalid mission type")
        return
    columns = [first_column, Text("Planet"), Text("System"), Sequence("Specials")]
    if show_detail:
        columns.append(Sequence("Detail"))
    candidate_table = Table(*columns, table_name="Potential Targets for %s in Turn %d" % (mission, fo.currentTurn()))
    for planet_id, score_tuple in candidates:
        if score_tuple[0] > 0.5:
            planet = universe.getPlanet(planet_id)
            entries = [
                get_first_column_value(score_tuple),
                planet,
                universe.getSystem(planet.systemID),
                planet.specials,
            ]
            if show_detail:
                entries.append(score_tuple[-1])
            candidate_table.add_row(*entries)
    info(candidate_table)


class OrbitalColonizationPlan:
    def __init__(self, target_id: int, source_id: int):
        """
        :param target_id: id of the target planet to colonize
        :param source_id: id of the planet which should build the colony base
        """
        self.target = target_id
        self.source = source_id
        self.base_enqueued = False
        self.fleet_id = INVALID_ID
        self.__score = 0
        self.__last_score_update = -1

    def assign_base(self, fleet_id: int) -> bool:
        """
        Assign an outpost base fleet to execute the plan.

        It is expected that the fleet consists of only that one outpost base.

        :return: True on success, False on failure
        """
        if self.base_assigned:
            warning("Assigned a base to a plan that was already assigned a base to.")
            return False
        # give orders to perform the mission
        target = TargetPlanet(self.target)
        fleet_mission = get_aistate().get_fleet_mission(fleet_id)
        fleet_mission.set_target(MissionType.ORBITAL_OUTPOST, target)
        self.fleet_id = fleet_id
        return True

    def enqueue_base(self) -> bool:
        """
        Enqueue the base according to the plan.

        :return: True on success, False on failure
        """
        if self.base_enqueued:
            warning("Tried to enqueue a base eventhough already done that.")
            return False

        # find the best possible base design for the source planet
        universe = fo.getUniverse()
        best_ship, _, _ = get_best_ship_info(PriorityType.PRODUCTION_ORBITAL_OUTPOST, self.source)
        if best_ship is None:
            warning("Can't find optimized outpost base design at %s" % (universe.getPlanet(self.source)))
            try:
                best_ship = next(
                    design
                    for design in fo.getEmpire().availableShipDesigns
                    if "SD_OUTPOST_BASE" == fo.getShipDesign(design).name
                )
                debug("Falling back to base design SD_OUTPOST_BASE")
            except StopIteration:
                # fallback design not available
                return False

        # enqueue the design at the source planet
        retval = fo.issueEnqueueShipProductionOrder(best_ship, self.source)
        debug(
            f"Enqueueing Outpost Base at {universe.getPlanet(self.source)} for {universe.getPlanet(self.target)} with result {retval}"
        )

        if not retval:
            warning("Failed to enqueue outpost base at %s" % universe.getPlanet(self.source))
            return False

        self.base_enqueued = True
        return True

    @property
    def base_assigned(self):
        if self.fleet_id == INVALID_ID:
            return False

        fleet = fo.getUniverse().getFleet(self.fleet_id)
        if fleet:
            return True

        debug("The fleet assigned to the OrbitalColonizationPlan doesn't exist anymore.")
        self.fleet_id = INVALID_ID
        return False

    @property
    def score(self):
        if self.__last_score_update != fo.currentTurn():
            self.__update_score()
        return self.__score

    def __update_score(self):
        planet_score = calculate_planet_colonization_rating(
            planet_id=self.target,
            mission_type=MissionType.OUTPOST,
            spec_name=None,
            detail=None,
        )
        for species in get_colony_builders():
            this_score = calculate_planet_colonization_rating(
                planet_id=self.target,
                mission_type=MissionType.COLONISATION,
                spec_name=species,
                detail=None,
            )
            planet_score = max(planet_score, this_score)
        self.__last_score_update = fo.currentTurn()
        self.__score = planet_score

    def is_valid(self) -> bool:
        """
        Check the colonization plan for validity, i.e. if it could be executed in the future.

        The plan is valid if it is possible to outpost the target planet
        and if the planet envisioned to build the outpost bases can still do so.
        """
        universe = fo.getUniverse()

        # make sure target is valid
        target = universe.getPlanet(self.target)
        if target is None or (not target.unowned) or target.speciesName:
            return False

        # make sure source is valid
        source = universe.getPlanet(self.source)
        if not (
            source
            and source.ownedBy(fo.empireID())
            and source.speciesName
            and fo.getSpecies(source.speciesName).canColonize
        ):
            return False

        # appears to be valid
        return True


class OrbitalColonizationManager:
    """
    The OrbitalColonizationManager handles orbital colonization for the AI.

    :type _colonization_plans: dict[int, OrbitalColonizationPlan]
    :type num_enqueued_bases: int
    """

    def __init__(self):
        self._colonization_plans = {}
        self.num_enqueued_bases = 0

    def get_targets(self) -> list[int]:
        """
        Return all planets for which an orbital colonization plan exists.
        """
        return list(self._colonization_plans.keys())

    def create_new_plan(self, target_id: int, source_id: int):
        """
        Create and keep track of a new colonization plan for a target planet.

        :param target_id: id of the target planet
        :param source_id: id of the planet which is supposed to build the base
        """
        if target_id in self._colonization_plans:
            warning("Already have a colonization plan for this planet. Doing nothing.")
            return
        self._colonization_plans[target_id] = OrbitalColonizationPlan(target_id, source_id)

    def turn_start_cleanup(self):  # noqa: C901
        universe = fo.getUniverse()
        # clean up invalid or finished plans
        for pid in list(self._colonization_plans.keys()):
            if not self._colonization_plans[pid].is_valid():
                del self._colonization_plans[pid]

        # parse the production queue and find bases which no longer have valid
        # targets (e.g. the planet was colonized already).
        self.num_enqueued_bases = 0
        unaccounted_plans = dict(self._colonization_plans)

        # Check which plans still have valid bases assigned (possibly interrupted by combat last turn)
        for pid in list(unaccounted_plans.keys()):
            if unaccounted_plans[pid].base_assigned:
                del unaccounted_plans[pid]

        # find enqueued bases which are no longer needed and dequeue those.
        items_to_dequeue = []
        aistate = get_aistate()
        for idx, element in enumerate(fo.getEmpire().productionQueue):
            if element.buildType != EmpireProductionTypes.BT_SHIP or element.turnsLeft == -1:
                continue

            role = aistate.get_ship_role(element.designID)
            if role != ShipRoleType.BASE_OUTPOST:
                continue

            self.num_enqueued_bases += 1
            # check if a target for this base remains
            original_target = next(
                (
                    target
                    for target, plan in unaccounted_plans.items()
                    if plan.source == element.locationID and plan.base_enqueued
                ),
                None,
            )
            if original_target:
                debug("Base built at %d still has its original target." % element.locationID)
                del unaccounted_plans[original_target]
                continue

            # the original target may be no longer valid but maybe there is another
            # orbital colonization plan which wasn't started yet and has the same source planet
            alternative_target = next(
                (target for target, plan in unaccounted_plans.items() if plan.source == element.locationID), None
            )
            if alternative_target:
                debug(
                    "Reassigning base built at %d to new target %d as old target is no longer valid"
                    % (element.locationID, alternative_target)
                )
                self._colonization_plans[alternative_target].base_enqueued = True
                del unaccounted_plans[alternative_target]
                continue

            # final try: unstarted plans with source in the same system
            target_system = universe.getSystem(universe.getPlanet(element.locationID).systemID)
            alternative_plan = next(
                (
                    plan
                    for target, plan in unaccounted_plans.items()
                    if plan.source in target_system.planetIDs and not plan.base_enqueued and not plan.base_assigned
                ),
                None,
            )
            if alternative_plan:
                debug(
                    "Reassigning base enqueued at %d to new plan with target %d. Previous source was %d"
                    % (element.locationID, alternative_plan.target, alternative_plan.source)
                )
                alternative_plan.source = element.locationID
                alternative_plan.base_enqueued = True
                del unaccounted_plans[alternative_plan.target]
                continue

            debug(
                "Could not find a target for the outpost base enqueued at %s" % universe.getPlanet(element.locationID)
            )
            items_to_dequeue.append(idx)

        # TODO: Stop Building for targets with now insufficient colonization score

        # delete last items first so that queue index of remaining items
        # does not have to be adjusted
        # TODO: Only pause the production if could become valid again
        items_to_dequeue.sort(reverse=True)
        for idx in items_to_dequeue:
            fo.issueDequeueProductionOrder(idx)
            self.num_enqueued_bases -= 1

    def build_bases(self):
        empire = fo.getEmpire()
        if not empire.techResearched(OUTPOSTING_TECH):
            return

        considered_plans = [
            plan
            for plan in self._colonization_plans.values()
            if not plan.base_enqueued and plan.score > MINIMUM_COLONY_SCORE
        ]
        queue_limit = max(1, int(2 * empire.productionPoints / outpod_pod_cost()))
        for colonization_plan in sorted(considered_plans, key=lambda x: x.score, reverse=True):
            if self.num_enqueued_bases >= queue_limit:
                debug("Base enqueue limit (%d) reached." % queue_limit)
                return

            success = colonization_plan.enqueue_base()
            if success:
                self.num_enqueued_bases += 1

    def assign_bases_to_colonize(self):
        universe = fo.getUniverse()
        all_outpost_base_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.ORBITAL_OUTPOST)
        avail_outpost_base_fleet_ids = FleetUtilsAI.extract_fleet_ids_without_mission_types(all_outpost_base_fleet_ids)
        for fid in avail_outpost_base_fleet_ids:
            fleet = universe.getFleet(fid)
            if not fleet:
                continue
            sys_id = fleet.systemID
            system = universe.getSystem(sys_id)

            avail_plans = [
                plan
                for plan in self._colonization_plans.values()
                if plan.target in system.planetIDs and not plan.base_assigned
            ]
            avail_plans.sort(key=lambda x: x.score, reverse=True)
            for plan in avail_plans:
                success = plan.assign_base(fid)
                if success:
                    break



from focs._effects import (
    EffectsGroup,
    Focus,
    Happiness,
    IsSource,
    NamedReal,
    OwnedBy,
    Planet,
    SetEmpireMeter,
    SetTargetIndustry,
    Source,
    TargetPopulation,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import TARGET_AFTER_2ND_SCALING_PRIORITY

Tech(
    name="PRO_SENTIENT_AUTOMATION",
    description="PRO_SENTIENT_AUTOMATION_DESC",
    short_description="INDUSTRY_AND_SLOT_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=800 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["LRN_PSIONICS", "PRO_ADAPTIVE_AUTOMATION"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & TargetPopulation(low=0.0001)
            & Focus(type=["FOCUS_INDUSTRY"])
            & Happiness(low=0),
            priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
            effects=SetTargetIndustry(value=Value + NamedReal(name="PRO_SENTIENT_AUTO_TARGET_INDUSTRY_FLAT", value=3)),
        ),
        EffectsGroup(
            scope=IsSource,
            effects=SetEmpireMeter(empire=Source.Owner, meter="ECONOMIC_CATEGORY_NUM_POLICY_SLOTS", value=Value + 1),
        ),
    ],
    graphic="icons/tech/basic_autofactories.png",
)



from focs._effects import (
    EffectsGroup,
    InSystem,
    IsBuilding,
    MaxOf,
    NamedReal,
    NamedRealLookup,
    Object,
    OwnedBy,
    Planet,
    Population,
    SetTargetHappiness,
    Source,
    SpeciesDislikes,
    SpeciesLikes,
    StatisticCount,
    Target,
    ThisBuilding,
    Value,
)
from macros.opinion import POLICY_DISLIKE_SCALING

STABILITY_PER_LIKED_BUILDING_ON_PLANET = 4.0


STABILITY_PER_LIKED_BUILDING_IN_SYSTEM = 1.0


STABILITY_PER_DISLIKED_BUILDING_ON_PLANET = 4.0 * POLICY_DISLIKE_SCALING

STABILITY_PER_DISLIKED_BUILDING_IN_SYSTEM = 1.0 * POLICY_DISLIKE_SCALING


SPECIES_LIKES_OR_DISLIKES_BUILDING_STABILITY_EFFECTS = [
    # species like building on the same planet
    EffectsGroup(
        scope=(Object(id=Source.PlanetID) & Population(low=0.001) & SpeciesLikes(name=ThisBuilding)),
        accountinglabel="LIKES_BUILDING_LABEL",
        effects=[SetTargetHappiness(value=Value + STABILITY_PER_LIKED_BUILDING_ON_PLANET)],
    ),
    # species dislike building on the same planet
    EffectsGroup(
        scope=(Object(id=Source.PlanetID) & Population(low=0.001) & SpeciesDislikes(name=ThisBuilding)),
        accountinglabel="DISLIKES_BUILDING_LABEL",
        effects=[SetTargetHappiness(value=Value - STABILITY_PER_DISLIKED_BUILDING_ON_PLANET)],
    ),
    # species like building in the same system
    EffectsGroup(
        scope=(
            Planet()
            & InSystem(id=Source.SystemID)
            & Population(low=0.001)
            & ~Object(id=Source.PlanetID)  # but not on same planet, which is covered by above case
            & SpeciesLikes(name=ThisBuilding)
        ),
        accountinglabel="LIKES_BUILDING_LABEL",
        effects=[SetTargetHappiness(value=Value + STABILITY_PER_LIKED_BUILDING_IN_SYSTEM)],
    ),
    # species dislike building in the same system
    EffectsGroup(
        scope=(
            Planet()
            & InSystem(id=Source.SystemID)
            & Population(low=0.001)
            & ~Object(id=Source.PlanetID)
            & SpeciesDislikes(name=ThisBuilding)
        ),
        accountinglabel="DISLIKES_BUILDING_LABEL",
        effects=[SetTargetHappiness(value=Value - STABILITY_PER_DISLIKED_BUILDING_IN_SYSTEM)],
    ),
    # species like building in empire
    EffectsGroup(
        scope=(
            Planet()
            & OwnedBy(empire=Source.Owner)
            & Population(low=0.001)
            & ~InSystem(id=Source.SystemID)
            & SpeciesLikes(name=ThisBuilding)
        ),
        accountinglabel="LIKES_BUILDING_LABEL",
        effects=[
            SetTargetHappiness(
                value=Value
                + (
                    NamedReal(name="BUILDING_LIKE_EMPIRE_SQRT_SCALING", value=0.5)
                    * MaxOf(
                        float,
                        1.0,
                        StatisticCount(
                            float,
                            condition=IsBuilding()
                            & IsBuilding(name=[ThisBuilding])
                            & OwnedBy(empire=Source.Owner)
                            & ~InSystem(id=Target.SystemID),
                        ),
                    )
                    ** (-0.5)
                )
            )
        ],
    ),
    # species dislike building in empire
    EffectsGroup(
        scope=(
            Planet()
            & OwnedBy(empire=Source.Owner)
            & Population(low=0.001)
            & ~InSystem(id=Source.SystemID)
            & SpeciesDislikes(name=ThisBuilding)
        ),
        accountinglabel="DISLIKES_BUILDING_LABEL",
        effects=[
            SetTargetHappiness(
                value=Value
                - (
                    NamedRealLookup(name="BUILDING_LIKE_EMPIRE_SQRT_SCALING")
                    * MaxOf(
                        float,
                        1.0,
                        StatisticCount(
                            float,
                            condition=IsBuilding()
                            & IsBuilding(name=[ThisBuilding])
                            & OwnedBy(empire=Source.Owner)
                            & ~InSystem(id=Target.SystemID),
                        ),
                    )
                    ** (-0.5)
                )
                * POLICY_DISLIKE_SCALING
            )
        ],
    ),
]



from common.listeners import register_post_handler
from handlers.shared_instances_code import (
    classes_to_exclude_from_universe,
    get_common_instances,
    get_item_with_location,
)
from stub_generator import generate_stub


def inspect_universe_generation_interface(*args, **kwargs):
    import freeorion as fo

    universe = fo.get_universe()

    # this field should be visible to AI
    empire_of_first_ai = fo.get_empire(2)  # first AI
    fo.create_field_in_system("FLD_NEBULA_1", 100, universe.getPlanet(empire_of_first_ai.capitalID).systemID)
    instances = list(get_item_with_location(get_common_instances()))

    generate_stub(
        fo,
        instances=instances,
        classes_to_ignore=classes_to_exclude_from_universe,
        path=".",
    )


register_post_handler("create_universe", inspect_universe_generation_interface)



from focs._species import *
from species.species_macros.detection import ULTIMATE_DETECTION
from species.species_macros.empire_opinions import FIXED_OPINION_EFFECTS
from species.species_macros.env import VERY_TOLERANT_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import NO_INDUSTRY
from species.species_macros.influence import ULTIMATE_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.population import GOOD_POPULATION
from species.species_macros.research import ULTIMATE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import ULTIMATE_STEALTH
from species.species_macros.stockpile import ULTIMATE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import ULTIMATE_DEFENSE_TROOPS

Species(
    name="SP_EXPERIMENTOR",
    description="SP_EXPERIMENTOR_DESC",
    gameplay_description="SP_EXPERIMENTOR_GAMEPLAY_DESC",
    tags=[
        "SELF_SUSTAINING",
        "NO_INDUSTRY",
        "ULTIMATE_INFLUENCE",
        "ULTIMATE_RESEARCH",
        "GOOD_POPULATION",
        "ULTIMATE_STOCKPILE",
        "ULTIMATE_DETECTION",
        "ULTIMATE_STEALTH",
        "PEDIA_SELF_SUSTAINING_SPECIES_CLASS",
    ],
    foci=[
        # [[HAS_INDUSTRY_FOCUS]]
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        # [[HAS_GROWTH_FOCUS]]
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "FRACTAL_GEODES_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "BLD_SHIPYARD_ENRG_COMP",
        "BLD_SHIPYARD_BASE",
        "BLD_SHIPYARD_ORBITAL_DRYDOCK",
        "BLD_BIOTERROR_PROJECTOR",
        "TIDAL_LOCK_SPECIAL",
        "ANCIENT_RUINS_DEPLETED_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "ELERIUM_SPECIAL",
    ],
    dislikes=[
        "BLD_SHIPYARD_ENRG_COMP",
        "BLD_CULTURE_ARCHIVES",
        "BLD_CULTURE_LIBRARY",
        "MONOPOLE_SPECIAL",
        "PANOPTICON_SPECIAL",
        "TEMPORAL_ANOMALY_SPECIAL",
        "SPICE_SPECIAL",
    ],
    annexation_condition=NoObject,
    effectsgroups=[
        NO_INDUSTRY,
        *ULTIMATE_RESEARCH,
        *ULTIMATE_INFLUENCE,
        *ULTIMATE_STOCKPILE,
        *GOOD_POPULATION,
        *AVERAGE_HAPPINESS,
        FIXED_OPINION_EFFECTS("SP_EXPERIMENTOR", 0),
        *AVERAGE_SUPPLY,
        *ULTIMATE_DEFENSE_TROOPS,
        *ULTIMATE_STEALTH,
        *ULTIMATE_DETECTION,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=VERY_TOLERANT_EP,
    graphic="icons/species/two-headed-03.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_ARCH_MONOFILS",
    description="CON_ARCH_MONOFILS_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=125 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=["CON_ASYMP_MATS"],
    unlock=Item(type=UnlockBuilding, name="BLD_SPACE_ELEVATOR"),
    graphic="icons/tech/architectural_monofilaments.png",
)



from freeorion_tools.design_compare import recursive_dict_diff


def test_compare_empty_dicts():
    assert recursive_dict_diff({}, {}, diff_level_threshold=1) == {}


def test_compare_dict_returns_intersection():
    dict_a = {1: 2, 2: {2: 3, 3: 4}}
    dict_b = {2: {2: 3, 3: 3}}
    assert recursive_dict_diff(dict_a, dict_b, diff_level_threshold=1) == {1: 2, 2: {3: 4}}



"""
Print utils.

For test proposes this module is not import any freeorion runtime libraries.
If you improve it somehow, add usage example to __main__ section.
"""
from collections import defaultdict
from collections.abc import Collection
from itertools import zip_longest
from math import ceil
from typing import Any, Union

from common.print_utils._base_field import Field


def as_columns(items: Collection[Any], columns=2) -> str:
    """
    Split flat list to columns and print them.

    >>> as_columns(['a', 'b', 'c', 'd'], 2)
    a   c
    b   d
    """
    row_count = int(ceil(len(items) / columns))
    text_columns = list(zip_longest(*[iter(items)] * row_count, fillvalue=""))
    column_widths = (max(len(x) for x in word) for word in text_columns)
    template = "   ".join("%%-%ss" % w for w in column_widths)

    return "\n".join(template % row for row in zip(*text_columns))


class Table:
    def __init__(
        self,
        *fields: Union[Field, Collection[Field]],
        vertical_sep="|",
        header_sep="=",
        bottom_sep="-",
        table_name="",
        hide_header=False,
    ):
        """
        Table layout for print data.

        - specify headers in constructor
        - add rows
        - print
        """
        self._table_name = table_name
        self._bottom_sep = bottom_sep
        self._header_sep = header_sep
        self._vertical_sep = vertical_sep
        self._rows = []
        self._notes = []  # each row has a not attached for it
        self._headers = fields
        self._hide_headers = hide_header
        self.totals = defaultdict(int)

    def __str__(self):
        return self.get_table()

    def add_row(self, *row, note: str = ""):
        table_row = []
        for field, val in zip(self._headers, row):
            if field.total:
                self.totals[field] += val
            table_row.append(field.make_cell_string(val))
        self._rows.append(table_row)
        self._notes.append(note)

    def _get_row_separator(self, char, column_widthes):
        return char * (2 + (len(column_widthes) - 1) * 3 + sum(column_widthes) + 2)

    def _get_rows_width(self) -> list[int]:
        if self._rows:
            return [max([len(y) for y in x], default=0) for x in zip(*self._rows)]
        else:
            return [1] * len(self._headers)

    def _get_headers_width(self) -> list[int]:
        if self._hide_headers:
            return [1] * len(self._headers)
        else:
            return [len(x.name) for x in self._headers]

    def _get_table_column_width(self) -> list[int]:
        columns_width = self._get_rows_width()
        header_width = self._get_headers_width()
        return [max(a, b) for a, b in zip(columns_width, header_width)]

    def __iter__(self):
        column_widths = self._get_table_column_width()

        if self._table_name:
            yield self._table_name

        inner_separator = " %s " % self._vertical_sep

        if not self._hide_headers:
            yield self._get_row_separator(self._header_sep, column_widths)

            yield "{} {} {}".format(
                self._vertical_sep,
                inner_separator.join(h.format_header(width) for h, width in zip(self._headers, column_widths)),
                self._vertical_sep,
            )

        yield self._get_row_separator(self._header_sep, column_widths)

        for row, note in zip(self._rows, self._notes):
            if note:
                note = f"  {note.strip()}"

            row_content = inner_separator.join(
                h.format_cell(item, width) for h, item, width in zip(self._headers, row, column_widths)
            )
            yield f"{self._vertical_sep} {row_content} {self._vertical_sep}{note}"

        if self.totals:
            yield self._get_row_separator(self._header_sep, column_widths)

            inner = inner_separator.join(
                h.format_cell(self.totals.get(h, " "), width) for h, width in zip(self._headers, column_widths)
            )

            yield f"{self._vertical_sep} {inner} {self._vertical_sep}"
            yield self._get_row_separator(self._header_sep, column_widths)
        else:
            yield self._get_row_separator(self._bottom_sep, column_widths)

        # print legend
        legend = [x for x in self._headers if x.description]
        if legend:
            name_width = max(len(x.name) for x in legend)
            for header in legend:
                yield "*%-*s %s" % (name_width, header.name[:-1], header.description)

    def get_table(self) -> str:
        """
        Return table as text.

        This method is deprecated, since long output will be truncated by logger.
        Use  print_table instead.
        """

        return "\n".join(self)



import freeorion as fo
import random
import sys

import universe_tables as tables
import util

# tuple of all valid planet sizes (with "no world")
planet_sizes_all = (
    fo.planetSize.tiny,
    fo.planetSize.small,
    fo.planetSize.medium,
    fo.planetSize.large,
    fo.planetSize.huge,
    fo.planetSize.asteroids,
    fo.planetSize.gasGiant,
    fo.planetSize.noWorld,
)

# tuple of planet sizes without "no world"
planet_sizes = (
    fo.planetSize.tiny,
    fo.planetSize.small,
    fo.planetSize.medium,
    fo.planetSize.large,
    fo.planetSize.huge,
    fo.planetSize.asteroids,
    fo.planetSize.gasGiant,
)

# tuple of "real" planet sizes (without "no world", asteroids and gas giants)
planet_sizes_real = (
    fo.planetSize.tiny,
    fo.planetSize.small,
    fo.planetSize.medium,
    fo.planetSize.large,
    fo.planetSize.huge,
)

# tuple of all available planet types (with asteroids and gas giants)
planet_types = (
    fo.planetType.swamp,
    fo.planetType.radiated,
    fo.planetType.toxic,
    fo.planetType.inferno,
    fo.planetType.barren,
    fo.planetType.tundra,
    fo.planetType.desert,
    fo.planetType.terran,
    fo.planetType.ocean,
    fo.planetType.asteroids,
    fo.planetType.gasGiant,
)

# tuple of available planet types without asteroids and gas giants ("real" planets)
planet_types_real = (
    fo.planetType.swamp,
    fo.planetType.radiated,
    fo.planetType.toxic,
    fo.planetType.inferno,
    fo.planetType.barren,
    fo.planetType.tundra,
    fo.planetType.desert,
    fo.planetType.terran,
    fo.planetType.ocean,
)


def base_chance_of_planet(planet_density, galaxy_shape, star_type, orbit, planet_size):
    return (
        tables.DENSITY_MOD_TO_PLANET_SIZE_DIST[planet_density][planet_size]
        + tables.STAR_TYPE_MOD_TO_PLANET_SIZE_DIST[star_type][planet_size]
        + tables.ORBIT_MOD_TO_PLANET_SIZE_DIST[orbit][planet_size]
        + tables.GALAXY_SHAPE_MOD_TO_PLANET_SIZE_DIST[galaxy_shape][planet_size]
    )


def can_have_planets(star_type, orbits, planet_density, galaxy_shape):
    """
    Return True if a system can have any planets.
    """
    for orbit in orbits:
        base_chance_none = base_chance_of_planet(planet_density, galaxy_shape, star_type, orbit, fo.planetSize.noWorld)
        base_chance_any = max(
            [base_chance_of_planet(planet_density, galaxy_shape, star_type, orbit, size) for size in planet_sizes]
        )

        if base_chance_any > base_chance_none - 100:
            return True
    return False


def calc_planet_size(star_type, orbit, planet_density, galaxy_shape):
    """
    Calculate planet size for a potential new planet based on planet density setup option, star type and orbit number.
    """
    # try to pick a planet size by making a series of "rolls" (1-100)
    # for each planet size, and take the highest modified roll
    planet_size = fo.planetSize.unknown
    try:
        max_roll = 0
        for size in planet_sizes_all:
            roll = random.randint(1, 100) + base_chance_of_planet(planet_density, galaxy_shape, star_type, orbit, size)
            if max_roll < roll:
                max_roll = roll
                planet_size = size
    except:  # noqa: E722
        # in case of an error play it safe and set planet size to invalid
        planet_size = fo.planetSize.unknown
        util.report_error("Python calc_planet_size: Pick planet size failed" + str(sys.exc_info()[1]))

    # if we got an invalid planet size (for whatever reason),
    # just select one randomly from the global tuple based
    # only on the planet density setup option
    if planet_size == fo.planetSize.unknown:
        if random.randint(1, 10) <= planet_density:
            planet_size = random.choice(planet_sizes)
        else:
            planet_size = fo.planetSize.noWorld

    return planet_size


def calc_planet_type(star_type, orbit, planet_size):
    """
    Calculate planet type randomly for a potential new planet.

    TODO: take into account star type and orbit number for determining planet type.
    """
    # check specified planet size to determine if we want a planet at all
    if planet_size in planet_sizes:
        # if yes, determine planet type based on planet size...
        if planet_size == fo.planetSize.gasGiant:
            return fo.planetType.gasGiant
        elif planet_size == fo.planetSize.asteroids:
            return fo.planetType.asteroids
        else:
            return random.choice(planet_types_real)
    else:
        return fo.planetType.unknown


def generate_a_planet(system, star_type, orbit, planet_density, galaxy_shape):
    """
    Place a planet in an orbit of a system. Return True on success
    """
    planet_size = calc_planet_size(star_type, orbit, planet_density, galaxy_shape)
    if planet_size not in planet_sizes:
        return False
    # ok, we want a planet, determine planet type and generate the planet
    planet_type = calc_planet_type(star_type, orbit, planet_size)
    if planet_type == fo.planetType.unknown:
        return False
    if fo.create_planet(planet_size, planet_type, system, orbit, "") == fo.invalid_object():
        # create planet failed, report an error
        util.report_error("Python generate_systems: create planet in system %d failed" % system)
        return False
    return True



from common.option_tools import TIMERS_USE_TIMERS, check_bool, get_option_dict
from common.timers import Timer

# setup module
options = get_option_dict()

try:
    USE_TIMERS = check_bool(options[TIMERS_USE_TIMERS])
except KeyError:
    USE_TIMERS = False


class DummyTimer:
    """
    Dummy timer to be used if timers are disabled.
    """

    def __init__(self, *args, **kwargs):
        pass

    def stop(self, *args, **kwargs):
        pass

    def start(self, *args, **kwargs):
        pass

    def stop_print_and_clear(self):
        pass


AITimer = Timer if USE_TIMERS else DummyTimer



from __future__ import annotations

from typing import Any, TypeVar, overload

from typing_extensions import TypeAlias

# Default value for int keyword arguments, use it when default value is not clear
from focs._types import (
    _ID,
    _T,
    _Aggregator,
    _BuildingType,
    _Condition,
    _ConditionalComposition,
    _DesignID,
    _Effect,
    _EffectGroup,
    _Empire,
    _EmpireId,
    _FloatParam,
    _Focus,
    _IntParam,
    _MeterType,
    _PlanetEnvironment,
    _PlanetId,
    _PlanetSize,
    _PlanetType,
    _Resource,
    _SpeciesValue,
    _StarType,
    _StringParam,
    _SystemID,
    _ValueParam,
    _Visibility,
)

Capacity = _MeterType()
MaxCapacity = _MeterType()
SecondaryStat = _MeterType()
MaxSecondaryStat = _MeterType()

Blue = _StarType()
White = _StarType()
Orange = _StarType()
Yellow = _StarType()
Red = _StarType()
Neutron = _StarType()
NoStar = _StarType()
BlackHole = _StarType()

def Object(id) -> _Condition: ...

Invisible = _Visibility()
Basic = _Visibility()
Partial = _Visibility()
Full = _Visibility()

Tiny = _PlanetSize()
Small = _PlanetSize()
Medium = _PlanetSize()
Large = _PlanetSize()
Huge = _PlanetSize()

Swamp = _PlanetType()
Toxic = _PlanetType()
Inferno = _PlanetType()
Radiated = _PlanetType()
Barren = _PlanetType()
Tundra = _PlanetType()
Desert = _PlanetType()
Terran = _PlanetType()
Ocean = _PlanetType()
AsteroidsType = _PlanetType()
GasGiantType = _PlanetType()

Good = _PlanetEnvironment()
Adequate = _PlanetEnvironment()
Poor = _PlanetEnvironment()
Hostile = _PlanetEnvironment()
Uninhabitable = _PlanetEnvironment()

ThisSpecies = _SpeciesValue()
ThisBuilding = _BuildingType()

BuildBuilding = _BuildingType()

AnyEmpire = _Empire()
EnemyOf = _Empire()
AllyOf = _Empire()
CurrentTurn = 0

ContentFocus = _Focus()

class _SystemInfo:
    LastTurnBattleHere: int
    ID: _ID

class Source:
    """
    FOCS Source condition is IsSource, this class is for value ref Source.<something>
    """

    Owner: _Empire
    ID: _ID
    SystemID: _SystemID
    Species: _SpeciesValue
    System: _SystemInfo
    PlanetID: _PlanetId
    TargetPopulation: float
    Research: float
    Industry: float
    HabitableSize: float
    CreationTurn: int

class LocalCandidate:
    LastTurnConquered: int
    LastTurnActiveInBattle: int
    LastTurnColonized: int
    TurnsSinceFocusChange: int
    LastTurnAttackedByShip: int
    LastTurnResupplied: int
    ETA: int
    ArrivedOnTurn: int
    ID: _ID
    Owner: _Empire
    Focus: _Focus
    NextSystemID: _SystemID
    System: _SystemInfo
    OrderedColonizePlanetID: _PlanetId
    Species: _SpeciesValue
    LastInvadedByEmpire: _Empire
    LastColonizedByEmpire: _Empire
    MaxStructure: float
    Industry: float
    TargetResearch: float
    Research: float
    TargetIndustry: float
    Stockpile: float
    MaxStockpile: float
    Construction: float
    TargetConstruction: float
    Influence: float
    TargetInfluence: float
    Population: float
    TargetPopulation: float
    Happiness: float
    TargetHappiness: float
    PlanetID: _PlanetId

class RootCandidate:
    ID: _ID
    Owner: _Empire
    PreviousSystemID: _SystemID

class Target:
    Owner: _Empire
    ID: _ID
    SystemID: _SystemID
    PlanetID: _PlanetId
    DesignID: _DesignID
    Construction: Any
    PreviousSystemID: _SystemID
    TurnsSinceFocusChange: int
    Population: float
    Happiness: float
    HabitableSize: float
    MaxSupply: float
    MaxFuel: float
    MaxStructure: float
    MaxTroops: float
    MaxDefense: float
    MaxShield: float
    MaxStockpile: float
    TargetPopulation: float
    TargetResearch: float
    TargetInfluence: float
    TargetIndustry: float
    TargetConstruction: float
    TargetHappiness: float

_O = TypeVar("_O", str, int, float)

class _ValuePlaceholder(_ConditionalComposition):
    def __call__(self, arg: _T) -> _T: ...
    def __add__(self, other: _T) -> _T: ...
    def __radd__(self, other: _T) -> _T: ...
    def __sub__(self, other: _T) -> _T: ...
    def __rsub__(self, other: _T) -> _T: ...
    def __mul__(self, other: _T) -> _T: ...
    def __rmul__(self, other: _T) -> _T: ...
    def __floordiv__(self, other: _T) -> _T: ...
    def __rfloordiv__(self, other: _T) -> _T: ...
    def __truediv__(self, other: _T) -> _T: ...
    def __rtruediv__(self, other: _T) -> _T: ...
    def __pow__(self, other: _T) -> _T: ...
    def __rpow__(self, other: _T) -> _T: ...

Value = _ValuePlaceholder()
ValueVisibility = _Visibility()

Min = _Aggregator()
Max = _Aggregator()
Sum = _Aggregator()
Mode = _Aggregator()

GalaxySize: _FloatParam = ...

_args_type = int | float | str | _ValuePlaceholder

def OneOf(type_: type[_T], *args: _args_type) -> _T: ...
def Abs(type_: type[_T], value: _args_type) -> _T: ...
def Floor(type_: type[_T], value: _FloatParam) -> _T: ...
@overload
def MinOf(type_: type[_T], *args: _args_type) -> _T: ...
@overload
def MinOf(type_: str, *args: _Visibility) -> int: ...
@overload
def MaxOf(type_: type[_T], *args: _args_type) -> _T: ...
@overload
def MaxOf(type_: str, *args: _Visibility) -> int: ...
def SpeciesColoniesOwned(*, empire: _Empire) -> int: ...
def DirectDistanceBetween(target: _ID, source: _ID) -> float: ...
def StatisticIf(type_: type[_T], condition: _Condition) -> _T: ...
def StatisticCount(type_: type[_T], condition: _Condition) -> _T: ...
def Statistic(type_: type[_T], aggregator: _Aggregator, value: _ValueParam, condition: _Condition) -> _T: ...
def NamedReal(*, name: str, value: _FloatParam) -> float: ...
def NamedRealLookup(*, name: str) -> float: ...
def NamedIntegerLookup(*, name: str) -> int: ...
def JumpsBetween(source: _ID, target: _ID) -> int: ...
def SpeciesShipsDestroyed(empire: _EmpireId, name: _SpeciesValue) -> int: ...
def SpeciesShipsLost(empire: _EmpireId, name: _SpeciesValue) -> int: ...
def SpeciesEmpireTargetOpinion(*, species: _SpeciesValue, empire: _EmpireId) -> float: ...
def EmpireStockpile(empire: _EmpireId, resource: _Resource) -> float: ...
def PartCapacity(*, name: str) -> float: ...
def PartSecondaryStat(*, name: str) -> float: ...
def PartsInShipDesign(*, name: str, design: _DesignID) -> float: ...
def ShipPartMeter(*, part: str, meter: _MeterType, ship: _ID): ...
def UserString(name: str) -> str: ...
def EmpireMeterValue(empire: _EmpireId, meter: str) -> float: ...
def EffectsGroup(
    *,
    scope: _Condition,
    effects: list[_Effect] | _Effect,
    activation: bool | _Condition | None = None,
    description: str = "",
    priority=None,
    accountinglabel: str = "",
    stackinggroup: str = "",
) -> _EffectGroup: ...

# Conditions

CurrentContent = ""

def Enqueued(type: _BuildingType, name: str) -> _Condition: ...
def TurnTechResearched(*, empire: _EmpireId, name: str) -> _Condition: ...
def IsBuilding(*, name: list[str | _BuildingType] = ...) -> _Condition: ...
def Number(*, low: int, condition: _Condition) -> _Condition: ...
def Planet(
    *, type: list[_PlanetType] = ..., environment: list[_PlanetEnvironment] = ..., size: list[_PlanetSize] = ...
) -> _Condition: ...
def Star(*, type: list[_StarType]) -> _Condition: ...
def Location(type: _Focus, name: _SpeciesValue, name2: _Focus | str) -> _Condition: ...
def HasSpecies(*, name: list[_StringParam] = ...) -> _Condition: ...
def IsField(*, name: list[_StringParam]) -> _Condition: ...
def Population(*, low: _FloatParam = ..., high: _FloatParam = ...) -> _Condition: ...
def TargetPopulation(*, low: _FloatParam = ..., high: _FloatParam = ...) -> _Condition: ...
def SpeciesLikes(*, name: _Focus | _BuildingType) -> _Condition: ...
def SpeciesDislikes(*, name: _Focus | _BuildingType) -> _Condition: ...
def Homeworld(*, name: list[_SpeciesValue] = ...) -> _Condition: ...
def HasTag(*, name: str) -> _Condition: ...
def VisibleToEmpire(*, empire: _EmpireId) -> _Condition: ...
def ContainedBy(condition: _Condition) -> _Condition: ...
def HasSpecial(*, name: str) -> _Condition: ...
def OwnerHasTech(*, name: str) -> _Condition: ...

System = _Condition()
Ship = _Condition()
Monster = _Condition()
Armed = _Condition()
Unowned = _Condition()
Capital = _Condition()
IsHuman = _Condition()
Fleet = _Condition()
Stationary = _Condition()
IsSource = _Condition()
IsTarget = _Condition()
CanColonize = _Condition()

def InSystem(*, id: _SystemID = _SystemID()) -> _Condition: ...

GalaxyMaxAIAggression = 0
GalaxyMonsterFrequency = 0

def MinimumNumberOf(number: int, sortkey: _FloatParam, condition: _Condition) -> _Condition: ...
def TargetIndustry(*, low: int) -> _Condition: ...
def Happiness(*, low: _FloatParam) -> _Condition: ...
def Focus(*, type: list[str]) -> _Condition: ...
def Random(*, probability: _FloatParam) -> _Condition: ...
def OwnedBy(*, affiliation: _EmpireId = _Empire(), empire: _EmpireId = _Empire()) -> _Condition: ...
def WithinStarlaneJumps(*, jumps: _IntParam, condition: _Condition) -> _Condition: ...
def WithinDistance(*, distance: int, condition: _Condition) -> _Condition: ...
def ResourceSupplyConnected(*, empire: _EmpireId, condition: _Condition) -> _Condition: ...
def Contains(scope: _Condition) -> _Condition: ...
def Turn(*, high: _IntParam = ..., low: _IntParam = ...) -> _Condition: ...
def Structure(*, low: _FloatParam = ..., high: _FloatParam = ...) -> _Condition: ...
def ResupplyableBy(*, empire: _EmpireId) -> _Condition: ...
def DesignHasPart(*, name: str, low: int = ..., high: int = ...) -> _Condition: ...
def EmpireHasAdoptedPolicy(*, name: str, empire: _EmpireId = _Empire()) -> _Condition: ...

Influence = _Resource()

def HasEmpireStockpile(*, empire: _EmpireId, resource: _Resource, low: int = ..., high: int = ...) -> _Condition: ...
def NumPoliciesAdopted(*, empire: _EmpireId) -> _Condition: ...
def OnPlanet(*, id: _ID) -> _Condition: ...

# Effects
Destroy = _Effect()

_RuleType: TypeAlias = type[_T]

def GameRule(
    *,
    type: _RuleType,
    name: str,
) -> float: ...

NoEffect = _Effect()

def Conditional(
    *,
    condition: _Condition | bool,
    effects: list[_Effect],
    else_: list[_Effect] = ...,
) -> _Effect: ...
def AddSpecial(*, name: str) -> _Effect: ...
def RemoveSpecial(*, name: str) -> _Effect: ...
def SetMaxShield(*, value: _FloatParam, accountinglabel: str = "") -> _Effect: ...
def SetMaxStructure(*, value: _FloatParam, accountinglabel: str = "") -> _Effect: ...
def SetEmpireStockpile(*, resource: _Resource, value: _FloatParam) -> _Effect: ...
def SetShield(*, value: _FloatParam) -> _Effect: ...
def SetTargetIndustry(*, value: _FloatParam) -> _Effect: ...
def SetDetection(*, value: _FloatParam) -> _Effect: ...
def SetStructure(*, value: _FloatParam) -> _Effect: ...
def SetFuel(*, value: _FloatParam) -> _Effect: ...
def SetDefense(*, value: _FloatParam) -> _Effect: ...
def SetMaxTroops(*, value: _FloatParam) -> _Effect: ...
def SetStockpile(*, value: _FloatParam) -> _Effect: ...
def SetMaxStockpile(*, value: _FloatParam, accountinglabel: str = "") -> _Effect: ...
def SetTargetHappiness(*, value: _FloatParam) -> _Effect: ...
def GiveEmpirePolicy(*, name: str, empire: _EmpireId = _Empire()) -> _Effect: ...
def SetVisibility(empire: _EmpireId, visibility: Any) -> _Effect: ...
def SetPopulation(*, value: _FloatParam) -> _Effect: ...
def SetPlanetType(*, type: _PlanetType) -> _Effect: ...
def SetSpecies(*, name: _StringParam) -> _Effect: ...
def SetTroops(*, value: _FloatParam) -> _Effect: ...
def SetMaxDefense(*, value: _FloatParam) -> _Effect: ...
def SetSupply(*, value: _FloatParam) -> _Effect: ...
def SetTargetResearch(*, value: _FloatParam) -> _Effect: ...
def SetHappiness(*, value: _FloatParam) -> _Effect: ...
def SetPlanetSize(*, planetsize: _PlanetSize) -> _Effect: ...
def SetMaxSupply(*, value: _FloatParam) -> _Effect: ...
def SetMaxFuel(*, value: _FloatParam) -> _Effect: ...
def SetResearch(*, value: _FloatParam) -> _Effect: ...
def SetIndustry(*, value: _FloatParam) -> _Effect: ...
def SetInfluence(*, value: _FloatParam) -> _Effect: ...
def SetSpeciesTargetOpinion(*, species: _SpeciesValue, empire: _EmpireId, opinion: _FloatParam) -> _Effect: ...
def SetSpeciesOpinion(*, species: _SpeciesValue, empire: _EmpireId, opinion: _FloatParam) -> _Effect: ...
def SetTargetPopulation(*, value: _FloatParam, accountinglabel: str = "") -> _Effect: ...
def SetTargetInfluence(*, value: _FloatParam) -> _Effect: ...
def SetTargetConstruction(*, value: _FloatParam) -> _Effect: ...
def SetConstruction(*, value: _FloatParam) -> _Effect: ...
def SetRebelTroops(*, value: _FloatParam) -> _Effect: ...
def SetStealth(*, value: _FloatParam) -> _Effect: ...
def CreateBuilding(*, type: str) -> _Effect: ...
def CreateShip(*, designname: str, species: str = "") -> _Effect: ...
def SetCapacity(*, partname: str, value: _FloatParam) -> _Effect: ...
def SetSecondaryStat(*, partname: str, value: _FloatParam) -> _Effect: ...
def SetMaxDamage(*, partname: str, value: _FloatParam) -> _Effect: ...
def SetMaxCapacity(*, partname: str, value: _FloatParam) -> _Effect: ...
def SetMaxSecondaryStat(*, partname: str, value: _FloatParam) -> _Effect: ...
def MoveTo(*, destination: _Condition) -> _Effect: ...
def MoveTowards(*, speed: int, target: _Condition) -> _Effect: ...
def SetStarType(*, type: _StarType) -> _Effect: ...
def SetEmpireMeter(*, empire: _EmpireId, meter: str, value: _FloatParam) -> _Effect: ...
def SetOwner(*, empire: _EmpireId) -> _Effect: ...
def GenerateSitRepMessage(
    *,
    message: str,
    label: str,
    empire: _EmpireId,
    parameters: dict[str, Any] = {},
    icon: str = "",
) -> _Effect: ...
def GiveEmpireTech(*, name: str, empire: _EmpireId = _Empire()) -> _Effect: ...
def Victory(*, reason: str) -> _Effect: ...



# BROAD_EP and NARROW_EP are stub placeholders for applying the
# board/narrow environment tolerance description to a species.
from focs._effects import (
    Adequate,
    AsteroidsType,
    Barren,
    Desert,
    GasGiantType,
    Good,
    Hostile,
    Inferno,
    Ocean,
    Poor,
    Radiated,
    Swamp,
    Terran,
    Toxic,
    Tundra,
    Uninhabitable,
)
from macros.misc import DESCRIPTION_EFFECTSGROUP_MACRO

BROAD_EP = DESCRIPTION_EFFECTSGROUP_MACRO("BROAD_EP_DESC")

NARROW_EP = DESCRIPTION_EFFECTSGROUP_MACRO("NARROW_EP_DESC")

SWAMP_STANDARD_EP = {
    Swamp: Good,
    Toxic: Adequate,
    Inferno: Poor,
    Radiated: Hostile,
    Barren: Hostile,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Poor,
    Ocean: Adequate,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

TOXIC_STANDARD_EP = {
    Swamp: Adequate,
    Toxic: Good,
    Inferno: Adequate,
    Radiated: Poor,
    Barren: Hostile,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Poor,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

INFERNO_STANDARD_EP = {
    Swamp: Poor,
    Toxic: Adequate,
    Inferno: Good,
    Radiated: Adequate,
    Barren: Poor,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

RADIATED_STANDARD_EP = {
    Swamp: Hostile,
    Toxic: Poor,
    Inferno: Adequate,
    Radiated: Good,
    Barren: Adequate,
    Tundra: Poor,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

BARREN_STANDARD_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Poor,
    Radiated: Adequate,
    Barren: Good,
    Tundra: Adequate,
    Desert: Poor,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

TUNDRA_STANDARD_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Poor,
    Barren: Adequate,
    Tundra: Good,
    Desert: Adequate,
    Terran: Poor,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

DESERT_STANDARD_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Poor,
    Tundra: Adequate,
    Desert: Good,
    Terran: Adequate,
    Ocean: Poor,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

TERRAN_STANDARD_EP = {
    Swamp: Poor,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Hostile,
    Tundra: Poor,
    Desert: Adequate,
    Terran: Good,
    Ocean: Adequate,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

OCEAN_STANDARD_EP = {
    Swamp: Adequate,
    Toxic: Poor,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Hostile,
    Tundra: Hostile,
    Desert: Poor,
    Terran: Adequate,
    Ocean: Good,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

SWAMP_NARROW_EP = {
    Swamp: Good,
    Toxic: Poor,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Hostile,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Poor,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

TOXIC_NARROW_EP = {
    Swamp: Poor,
    Toxic: Good,
    Inferno: Poor,
    Radiated: Hostile,
    Barren: Hostile,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

INFERNO_NARROW_EP = {
    Swamp: Hostile,
    Toxic: Poor,
    Inferno: Good,
    Radiated: Poor,
    Barren: Hostile,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

RADIATED_NARROW_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Poor,
    Radiated: Good,
    Barren: Poor,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

BARREN_NARROW_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Poor,
    Barren: Good,
    Tundra: Poor,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

TUNDRA_NARROW_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Poor,
    Tundra: Good,
    Desert: Poor,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

DESERT_NARROW_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Hostile,
    Tundra: Poor,
    Desert: Good,
    Terran: Poor,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

TERRAN_NARROW_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Hostile,
    Tundra: Hostile,
    Desert: Poor,
    Terran: Good,
    Ocean: Poor,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

OCEAN_NARROW_EP = {
    Swamp: Poor,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Hostile,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Poor,
    Ocean: Good,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

SWAMP_BROAD_EP = {
    Swamp: Good,
    Toxic: Adequate,
    Inferno: Poor,
    Radiated: Poor,
    Barren: Hostile,
    Tundra: Hostile,
    Desert: Poor,
    Terran: Poor,
    Ocean: Adequate,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

TOXIC_BROAD_EP = {
    Swamp: Adequate,
    Toxic: Good,
    Inferno: Adequate,
    Radiated: Poor,
    Barren: Poor,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Poor,
    Ocean: Poor,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

INFERNO_BROAD_EP = {
    Swamp: Poor,
    Toxic: Adequate,
    Inferno: Good,
    Radiated: Adequate,
    Barren: Poor,
    Tundra: Poor,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Poor,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

RADIATED_BROAD_EP = {
    Swamp: Poor,
    Toxic: Poor,
    Inferno: Adequate,
    Radiated: Good,
    Barren: Adequate,
    Tundra: Poor,
    Desert: Poor,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

BARREN_BROAD_EP = {
    Swamp: Hostile,
    Toxic: Poor,
    Inferno: Poor,
    Radiated: Adequate,
    Barren: Good,
    Tundra: Adequate,
    Desert: Poor,
    Terran: Poor,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

TUNDRA_BROAD_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Poor,
    Radiated: Poor,
    Barren: Adequate,
    Tundra: Good,
    Desert: Adequate,
    Terran: Poor,
    Ocean: Poor,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

DESERT_BROAD_EP = {
    Swamp: Poor,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Poor,
    Barren: Poor,
    Tundra: Adequate,
    Desert: Good,
    Terran: Adequate,
    Ocean: Poor,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

TERRAN_BROAD_EP = {
    Swamp: Poor,
    Toxic: Poor,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Poor,
    Tundra: Poor,
    Desert: Adequate,
    Terran: Good,
    Ocean: Adequate,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

OCEAN_BROAD_EP = {
    Swamp: Adequate,
    Toxic: Poor,
    Inferno: Poor,
    Radiated: Hostile,
    Barren: Hostile,
    Tundra: Poor,
    Desert: Poor,
    Terran: Adequate,
    Ocean: Good,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

TOLERANT_EP = {
    Swamp: Good,
    Toxic: Good,
    Inferno: Good,
    Radiated: Good,
    Barren: Good,
    Tundra: Good,
    Desert: Good,
    Terran: Good,
    Ocean: Good,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

VERY_TOLERANT_EP = {
    Swamp: Good,
    Toxic: Good,
    Inferno: Good,
    Radiated: Good,
    Barren: Good,
    Tundra: Good,
    Desert: Good,
    Terran: Good,
    Ocean: Good,
    AsteroidsType: Adequate,
    GasGiantType: Adequate,
}

ASTEROIDAL_NARROW_EP = {
    Swamp: Uninhabitable,
    Toxic: Uninhabitable,
    Inferno: Uninhabitable,
    Radiated: Uninhabitable,
    Barren: Hostile,
    Tundra: Uninhabitable,
    Desert: Uninhabitable,
    Terran: Uninhabitable,
    Ocean: Uninhabitable,
    AsteroidsType: Adequate,
    GasGiantType: Uninhabitable,
}

ASTEROIDAL_STANDARD_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Poor,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Good,
    GasGiantType: Uninhabitable,
}

GASEOUS_NARROW_EP = {
    Swamp: Uninhabitable,
    Toxic: Uninhabitable,
    Inferno: Uninhabitable,
    Radiated: Uninhabitable,
    Barren: Uninhabitable,
    Tundra: Uninhabitable,
    Desert: Uninhabitable,
    Terran: Uninhabitable,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Adequate,
}

GASEOUS_STANDARD_EP = {
    Swamp: Hostile,
    Toxic: Hostile,
    Inferno: Hostile,
    Radiated: Hostile,
    Barren: Hostile,
    Tundra: Hostile,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Hostile,
    AsteroidsType: Uninhabitable,
    GasGiantType: Good,
}



from freeorion_tools._fleet_position import get_fleet_position
from freeorion_tools._profile import profile
from freeorion_tools.combine_ratings import combine_ratings

from ._freeorion_tools import (
    ReadOnlyDict,
    assertion_fails,
    chat_human,
    get_ai_tag_grade,
    get_game_rule_int,
    get_game_rule_real,
    get_named_int,
    get_named_real,
    get_partial_visibility_turn,
    get_ship_part,
    get_species_attack_troops,
    get_species_fuel,
    get_species_industry,
    get_species_influence,
    get_species_population,
    get_species_research,
    get_species_ship_shields,
    get_species_stability,
    get_species_stealth,
    get_species_supply,
    get_species_tag_grade,
    policy_is_adopted,
    ppstring,
    tech_is_complete,
    tech_soon_available,
)



from stub_generator.interface_inspector import (
    ClassInfo,
    EnumInfo,
    FunctionInfo,
    InstanceInfo,
)
from stub_generator.stub_generator.class_generator import generate_classes
from stub_generator.stub_generator.enum_generator import generate_enums
from stub_generator.stub_generator.function_generator import generate_functions
from stub_generator.stub_generator.result_builder import Import, ResultBuilder


def make_stub(
    classes: list[ClassInfo],
    enums: list[EnumInfo],
    functions: list[FunctionInfo],
    instances: list[InstanceInfo],
    result_path,
    classes_to_ignore: set[str],
):
    header = (
        "# Autogenerated do not modify manually!\n"
        "# This is a type-hinting python stub file, used by python IDEs to provide type hints. For more information\n"
        "# about stub files, see https://www.python.org/dev/peps/pep-0484/#stub-files\n"
        "# During execution, the actual module is made available via\n"
        "# a C++ Boost-python process as part of the launch."
    )

    res = ResultBuilder(header)
    res.add_built_in_import(Import("collections.abc", ["Mapping", "Sequence", "Set"]))
    res.add_built_in_import(Import("enum", ["IntEnum"]))
    res.add_built_in_import(Import("typing", ["overload"]))

    res.add_import(
        Import(
            "common.fo_typing",
            [
                "BuildingId",
                "BuildingName",
                "EmpireId",
                "FleetId",
                "NamedTuple",
                "ObjectId",
                "PartName",
                "PlanetId",
                "PlayerId",
                "ShipId",
                "SpeciesName",
                "SystemId",
                "Turn",
            ],
        )
    )

    res.add_classes(generate_classes(classes, instances, classes_to_ignore, enums))
    res.add_enums(generate_enums(enums))
    res.add_functions(generate_functions(functions))

    with open(result_path, "w") as f:
        res.write(f)



"""
This module provides an interface for a quasi-singleton AIstate instance.

The AIstate must be initialized after module imports using either
   * create_new_aistate
   * load_aistate
This should happen only once per session and be triggered by the game client.

Access to the AIstate is provided via get_aistate().

Example usage:
    from aistate_interface import get_aistate()
    get_aistate().get_all_fleet_missions()

"""
_aistate = None


def create_new_aistate(*args, **kwargs):
    import AIstate

    global _aistate
    _aistate = AIstate.AIstate(*args, **kwargs)
    return _aistate


def load_aistate(savegame_string: str):
    import savegame_codec

    global _aistate
    _aistate = savegame_codec.load_savegame_string(savegame_string)
    return _aistate


def get_aistate():
    """
    :rtype: AIstate.AIstate
    """
    return _aistate



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_BLACKSHIELD",
    description="SHP_BLACKSHIELD_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_PARTS_CATEGORY",
    researchcost=2500 * TECH_COST_MULTIPLIER,
    researchturns=10,
    tags=["PEDIA_SHIELD_PART_TECHS"],
    prerequisites=["SHP_PLASMA_SHIELD"],
    unlock=Item(type=UnlockShipPart, name="SH_BLACK"),
    graphic="icons/ship_parts/blackshield.png",
)



from focs._effects import (
    BlackHole,
    Blue,
    EffectsGroup,
    Focus,
    Happiness,
    InSystem,
    NamedReal,
    Neutron,
    Orange,
    OwnedBy,
    Planet,
    Red,
    SetTargetResearch,
    Source,
    Star,
    StatisticCount,
    Target,
    Value,
    White,
    Yellow,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import TARGET_AFTER_SCALING_PRIORITY

Tech(
    name="LRN_STELLAR_TOMOGRAPHY",
    description="LRN_STELLAR_TOMOGRAPHY_DESC",
    short_description="RESEARCH_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=180 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_EVERYTHING"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & Star(type=[BlackHole])
            & Focus(type=["FOCUS_RESEARCH"])
            & Happiness(low=0),
            priority=TARGET_AFTER_SCALING_PRIORITY,
            effects=SetTargetResearch(
                value=Value
                + StatisticCount(
                    float,
                    condition=Planet()
                    & OwnedBy(empire=Source.Owner)
                    & InSystem(id=Target.SystemID)
                    & Focus(type=["FOCUS_RESEARCH"]),
                )
                * NamedReal(name="LRN_STELLAR_TOMO_BLACK_TARGET_RESEARCH_PERPLANET", value=3.0)
            ),
        ),
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & Star(type=[Neutron])
            & Focus(type=["FOCUS_RESEARCH"])
            & Happiness(low=0),
            priority=TARGET_AFTER_SCALING_PRIORITY,
            effects=SetTargetResearch(
                value=Value
                + StatisticCount(
                    float,
                    condition=Planet()
                    & OwnedBy(empire=Source.Owner)
                    & InSystem(id=Target.SystemID)
                    & Focus(type=["FOCUS_RESEARCH"]),
                )
                * NamedReal(name="LRN_STELLAR_TOMO_NEUTRON_TARGET_RESEARCH_PERPLANET", value=1.0)
            ),
        ),
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & Star(type=[Blue, White, Red, Orange, Yellow])
            & Focus(type=["FOCUS_RESEARCH"])
            & Happiness(low=0),
            priority=TARGET_AFTER_SCALING_PRIORITY,
            effects=SetTargetResearch(
                value=Value
                + StatisticCount(
                    float,
                    condition=Planet()
                    & OwnedBy(empire=Source.Owner)
                    & InSystem(id=Target.SystemID)
                    & Focus(type=["FOCUS_RESEARCH"]),
                )
                * NamedReal(name="LRN_STELLAR_TOMO_NORMAL_STAR_TARGET_RESEARCH_PERPLANET", value=0.2)
            ),
        ),
    ],
    graphic="icons/tech/stellar_tomography.png",
)



from focs._species import *
from macros.misc import DESCRIPTION_EFFECTSGROUP_MACRO
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import TERRAN_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import GOOD_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS

Species(
    name="SP_HUMAN",
    description="SP_HUMAN_DESC",
    gameplay_description="SP_HUMAN_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=["ORGANIC", "STYLISH", "ARTISTIC", "GOOD_INFLUENCE", "PEDIA_ORGANIC_SPECIES_CLASS", "PEDIA_ARTISTIC"],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "PLC_BOOTSTRAPPING",
        "PLC_MODERATION",
        "PLC_PROPAGANDA",
        "PLC_TRAFFIC_CONTROL",
        "FOCUS_INFLUENCE",
        "FOCUS_RESEARCH",
        "WORLDTREE_SPECIAL",
        "SHIMMER_SILK_SPECIAL",
        "FRACTAL_GEODES_SPECIAL",
        "ANCIENT_RUINS_DEPLETED_SPECIAL",
        "GAIA_SPECIAL",
    ],
    dislikes=[
        "PLC_AUGMENTATION",
        "PLC_CENTRALIZATION",
        "PLC_TECHNOCRACY",
        "PLC_THE_HUNT",
        "TIDAL_LOCK_SPECIAL",
        "FOCUS_BIOTERROR",
        "FOCUS_PROTECTION",
        "BLD_BIOTERROR_PROJECTOR",
        "BLD_SPATIAL_DISTORT_GEN",
        "BLD_SHIPYARD_ORG_XENO_FAC",
        "FORTRESS_SPECIAL",
        "TEMPORAL_ANOMALY_SPECIAL",
        "KRAKEN_NEST_SPECIAL",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *AVERAGE_RESEARCH,
        *GOOD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_HUMAN"),
        *AVERAGE_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        DESCRIPTION_EFFECTSGROUP_MACRO("AVERAGE_INDUSTRY_DESC"),
        DESCRIPTION_EFFECTSGROUP_MACRO("AVERAGE_RESEARCH_DESC"),
        DESCRIPTION_EFFECTSGROUP_MACRO("AVERAGE_STOCKPILE_DESC"),
        DESCRIPTION_EFFECTSGROUP_MACRO("AVERAGE_POPULATION_DESC"),
        DESCRIPTION_EFFECTSGROUP_MACRO("AVERAGE_SUPPLY_DESC"),
        DESCRIPTION_EFFECTSGROUP_MACRO("AVERAGE_FUEL_DESC"),
        DESCRIPTION_EFFECTSGROUP_MACRO("AVERAGE_DEFENSE_TROOPS_DESC"),
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=TERRAN_STANDARD_EP,
    graphic="icons/species/human.png",
)



from focs._effects import (
    Adequate,
    EffectsGroup,
    HasTag,
    Hostile,
    OwnedBy,
    Planet,
    Poor,
    SetTargetPopulation,
    Source,
    StatisticIf,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import TARGET_POPULATION_BEFORE_SCALING_PRIORITY

Tech(
    name="GRO_XENO_GENETICS",
    description="GRO_XENO_GENETICS_DESC",
    short_description="POPULATION_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=(
        120
        * TECH_COST_MULTIPLIER
        / (1 + StatisticIf(float, condition=OwnedBy(empire=Source.Owner) & HasTag(name="ADAPTIVE")))
    ),
    researchturns=8,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=[
        "GRO_SYMBIOTIC_BIO",
        "GRO_GENETIC_MED",
    ],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & Planet(environment=[Adequate, Poor]),
            accountinglabel="GRO_TECH_ACCOUNTING_LABEL",
            priority=TARGET_POPULATION_BEFORE_SCALING_PRIORITY,
            effects=SetTargetPopulation(value=Value + 2 * Target.HabitableSize),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & Planet(environment=[Hostile]),
            accountinglabel="GRO_TECH_ACCOUNTING_LABEL",
            priority=TARGET_POPULATION_BEFORE_SCALING_PRIORITY,
            effects=SetTargetPopulation(value=Value + 1 * Target.HabitableSize),
        ),
    ],
    graphic="icons/tech/xenological_genetics.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_ENDOCRINE_SYSTEMS",
    description="SHP_ENDOCRINE_SYSTEMS_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=80 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_ORGANIC_HULL_TECHS"],
    prerequisites=["SHP_MULTICELL_CAST", "GRO_LIFECYCLE_MAN"],
    unlock=[Item(type=UnlockShipHull, name="SH_ENDOMORPHIC")],
    graphic="icons/ship_hulls/endomorphic_hull_small.png",
)



DIFF_ADD = "#36CF70"
DIFF_REMOVE = "#A32929"


def to_hex_color(color):
    r, g, b = color["R"], color["G"], color["B"]
    return f"#{r:02X}{g:02X}{b:02X}"


def span_with_hint(text, hint):
    return f'<span title="{hint}">{text}</span>'


def colored_span(text, hex_color):
    return f'<span style="color:{hex_color}">{text}</span>'



from abc import abstractmethod
from typing import Any


class ChatFormatter:
    """
    Helper class for a message formatting.
    """

    @abstractmethod
    def _wrap_to_tag(self, tag: (str, str), message: Any) -> str:
        open_, close = tag
        return f"<{open_}>{message}</{close}>"

    @abstractmethod
    def _get_tag_rgb(self, color: (int, int, int, int)) -> (str, str):
        return "rgba {} {} {} {}".format(*color), "rgba"

    @abstractmethod
    def _get_underline_tag(self) -> (str, str):
        return "u", "u"

    def colored(self, color: (int, int, int, int), message: Any):
        return self._wrap_to_tag(self._get_tag_rgb(color), message)

    def underline(self, message: Any) -> str:
        return self._wrap_to_tag(self._get_underline_tag(), message)

    def white(self, message: Any) -> str:
        return self.colored((255, 255, 255, 255), message)

    def red(self, message: Any) -> str:
        return self.colored((255, 0, 0, 255), message)

    def blue(self, message: Any) -> str:
        return self.colored((0, 255, 255, 255), message)

    def yellow(self, message: Any) -> str:
        return self.colored((255, 255, 0, 255), message)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import TERRAN_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
)
from species.species_macros.happiness import BAD_HAPPINESS
from species.species_macros.industry import BAD_INDUSTRY
from species.species_macros.influence import NO_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import NO_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import BAD_SUPPLY
from species.species_macros.troops import ULTIMATE_DEFENSE_TROOPS

Species(
    name="SP_RAAAGH",
    description="SP_RAAAGH_DESC",
    gameplay_description="SP_RAAAGH_GAMEPLAY_DESC",
    native=True,
    tags=[
        "ORGANIC",
        "BAD_INDUSTRY",
        "NO_RESEARCH",
        "NO_INFLUENCE",
        "BAD_HAPPINESS",
        "BAD_SUPPLY",
        "PRIMITIVE",
        "PEDIA_ORGANIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_PROTECTION",
    likes=[
        "FOCUS_PROTECTION",
        "MIMETIC_ALLOY_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "BLD_BLACK_HOLE_POW_GEN",
        "HONEYCOMB_SPECIAL",
        "COMPUTRONIUM_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "FRUIT_SPECIAL",
        "PLC_CONFEDERATION",
        "PLC_CONFORMANCE",
        "PLC_DIVINE_AUTHORITY",
        "PLC_DREAM_RECURSION",
    ],
    dislikes=[
        "FOCUS_INDUSTRY",
        "BLD_GAS_GIANT_GEN",
        "BLD_AUTO_HISTORY_ANALYSER",
        "BLD_INTERSPECIES_ACADEMY",
        "BLD_GENOME_BANK",
        "BLD_PLANET_CLOAK",
        "BLD_SHIPYARD_CON_GEOINT",
        "BLD_SHIPYARD_ORG_ORB_INC",
        "PANOPTICON_SPECIAL",
        "SPICE_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "PLC_INTERSTELLAR_INFRA",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *BAD_INDUSTRY,
        NO_RESEARCH,
        *NO_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *BAD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_RAAAGH"),
        *BAD_SUPPLY,
        *ULTIMATE_DEFENSE_TROOPS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=TERRAN_STANDARD_EP,
    spawnrate=1,
    spawnlimit=1,
    graphic="icons/species/raaagh.png",
)



import streamlit as st

from log_parser.extract_data import process_state
from log_parser.file_reader import return_file_list
from log_parser.log_tokenizer import tokenize_log


def get_ais_data():
    timestamp, file_list = return_file_list()
    return _get_ais_data(timestamp, file_list)


@st.cache
def _get_ais_data(_cache_key, file_list: tuple):
    data = (process_state(tokenize_log(file_)) for file_ in file_list)
    return sorted(data, key=lambda x: int(x["empire_id"]))



from common.statistic_interface import StatKey


def start_turn(key, value, data):
    data["turns"].append({})


def add_attribute(key: StatKey, value, data):
    info = data["turns"][-1]
    if key.is_multi():
        info.setdefault(key.name, []).append(value)
    else:
        info[key.name] = value


def start_first_turn(key, value, data):
    start_turn(key, value, data)
    data["empire_id"] = value["empire_id"]
    data["player_name"] = value["name"]


def set_color(key, value, data):
    data["color"] = value


def process_output(key, value, data):
    info = data["turns"][-1]
    info["turn"] = value["turn"]
    info["PP"] = value["PP"]
    info["RP"] = value["RP"]


def skip(key, value, data):
    pass


DEFAULT = object()

states = {
    "new": {StatKey.EmpireID: (start_first_turn, "first_turn")},
    "first_turn": {
        StatKey.EmpireColors: (set_color, "turn"),
    },
    "turn": {
        StatKey.EmpireID: (start_turn, "turn"),
        StatKey.EmpireColors: (skip, "turn"),
        DEFAULT: (add_attribute, "turn"),
        StatKey.Output: (process_output, "turn"),
    },
}


def process_state(lines):
    data = {"turns": []}

    state = "new"
    for key, value in lines:
        handlers = states[state]
        if key in handlers:
            handler, state = handlers[key]
        else:
            handler, state = handlers[DEFAULT]
        handler(key, key.value.deserialize(value), data)
    return data



from focs._effects import (
    CurrentTurn,
    EffectsGroup,
    Focus,
    IsSource,
    LocalCandidate,
    MinOf,
    NamedRealLookup,
    Planet,
    SetDefense,
    SetMaxDefense,
    SetTargetHappiness,
    Target,
    Unowned,
    Value,
)
from macros.misc import PLANET_DEFENSE_FACTOR
from macros.priorities import (
    AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    TARGET_AFTER_2ND_SCALING_PRIORITY,
)

AVERAGE_PLANETARY_DEFENSE = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_PROTECTION"]),
        stackinggroup="FOCUS_PROTECTION_DEFENSE_STACK",
        accountinglabel="FOCUS_PROTECTION_LABEL",
        priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
        effects=[
            SetMaxDefense(value=Value * 2),
            SetTargetHappiness(value=Value + NamedRealLookup(name="PROTECION_FOCUS_STABILITY_BONUS")),
        ],
    ),
    EffectsGroup(  # increase 1 per turn, up to max
        scope=IsSource,
        activation=Planet() & Unowned & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
        accountinglabel="DEF_ROOT_DEFENSE",
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetDefense(value=MinOf(float, Value(Target.MaxDefense), Value + PLANET_DEFENSE_FACTOR)),
    ),
]


def NATIVE_PLANETARY_DEFENSE(defence: float):
    return EffectsGroup(
        scope=IsSource,
        activation=Planet() & Unowned,
        accountinglabel="NATIVE_PLANETARY_DEFENSE_LABEL",
        effects=SetMaxDefense(value=Value + defence),
    )



"""This module defines the encoding for the FreeOrion AI savegames.

The encoding is json-based with custom prefixes to support some objects
and dictionary keys of types which are not supported in standard json.

This module encodes the integer as string with prefix so that the information
about its integer key will be kept throughout json decoding.

The encoding implementation is recursive
    1) Identify the type of object to encode and call the correct encoder
    2) If it is a non-trivial type, first encode all its content
    3) Finally, encode the object itself

For class instances, the __getstate__ method is invoked to get its content.
If not defined, its __dict__ will be encoded instead.

If an object could not be encoded, raise a CanNotSaveGameException.
"""
import base64
import collections
import json
import zlib
from enum import IntEnum
from typing import Any, Callable

from ._definitions import (
    ENUM_PREFIX,
    FALSE,
    FLOAT_PREFIX,
    INT_PREFIX,
    NONE,
    PLACEHOLDER,
    SET_PREFIX,
    TRUE,
    TUPLE_PREFIX,
    CanNotSaveGameException,
    TrustedClasses,
)


def _encode_with_prefix(prefix, value):
    return f'"{prefix}{value}"'


def build_savegame_string() -> bytes:
    """Encode the AIstate and compress the resulting string with zlib.

    To decode the string, first call zlib.decompress() on it.

    :return: compressed savegame string
    """
    from aistate_interface import get_aistate

    trusted_classes = TrustedClasses()
    savegame_string = encode(get_aistate(), trusted_classes=trusted_classes)
    return base64.b64encode(zlib.compress(savegame_string.encode("utf-8")))


def encode(o: Any, trusted_classes: TrustedClasses) -> str:
    """Encode the passed object as json-based string.

    :param o: object to be encoded
    :return: String representation of the object state
    """
    o_type = type(o)

    # Find and call the correct encoder based
    # on the type of the object to encode

    type_encoder = TypeEncoder(trusted_classes)
    try:
        encoder_function = type_encoder.get_encoder_function(o_type)
    except KeyError:
        if issubclass(o_type, IntEnum):
            return _encode_with_prefix(ENUM_PREFIX, f"{o.__class__.__name__}.{o.name}")
        else:
            return type_encoder._encode_object(o)
    else:
        # only call now to not catch KeyError withing the encoder call
        return encoder_function(o)


class TypeEncoder:
    def __init__(self, trusted_classes: TrustedClasses):
        self._trusted_classes = trusted_classes

    def _encode_str(self, o):
        """
        Encode string.

        We don't check if prefixes for types (eg __ENUM__) is in the string.
        The string will be encoded, but it will not be decoded back.
        Since encoded object is under control of developers team,
         and it is possible to fix it manually by editing save game, we will leave it as is.
        """
        return json.dumps(o)

    def _encode_dict(self, o):
        """Get a string representation of a dict with its encoded content."""
        return "{%s}" % (
            ", ".join([f"{encode(k, self._trusted_classes)}: {encode(v, self._trusted_classes)}" for k, v in o.items()])
        )

    def _encode_set(self, o):
        """Get a string representation of a set with its encoded content."""
        return _encode_with_prefix(SET_PREFIX, "(%s)" % (self._encode_list(list(o)).replace('"', PLACEHOLDER)))

    def _encode_list(self, o):
        """Get a string representation of a list with its encoded content."""
        return "[%s]" % (", ".join([encode(v, self._trusted_classes) for v in o]))

    def _encode_tuple(self, o):
        """Get a string representation of a tuple with its encoded content."""
        return _encode_with_prefix(TUPLE_PREFIX, "(%s)" % (self._encode_list(list(o)).replace('"', PLACEHOLDER)))

    def _encode_none(self, o):
        return _encode_with_prefix(NONE, "")

    def _encode_int(self, o):
        return _encode_with_prefix(INT_PREFIX, str(o))

    def _encode_float(self, o):
        return _encode_with_prefix(FLOAT_PREFIX, repr(o))

    def _encode_bool(self, o):
        return _encode_with_prefix(TRUE, "") if o else _encode_with_prefix(FALSE, "")

    def _encode_object(self, obj):
        """Get a string representation of state of an object which is not handled by a specialized encoder."""
        try:
            class_name = f"{obj.__class__.__module__}.{obj.__class__.__name__}"
            if class_name not in self._trusted_classes:
                raise CanNotSaveGameException("Class %s is not trusted" % class_name)
        except AttributeError:
            # obj does not have a class or class has no module
            raise CanNotSaveGameException(f"Encountered unsupported object {obj} ({type(obj)})")

        # if possible, use getstate method to query the state, otherwise use the object's __dict__
        try:
            getstate = obj.__getstate__
        except AttributeError:
            value = obj.__dict__
        else:
            # only call now to avoid catching exceptions raised during the getstate call
            value = getstate()

        # encode information about class
        value.update({"__class__": obj.__class__.__name__, "__module__": obj.__class__.__module__})
        return self._encode_dict(value)

    def get_encoder_function(self, o) -> Callable:
        return {
            str: self._encode_str,
            bool: self._encode_bool,
            int: self._encode_int,
            float: self._encode_float,
            dict: self._encode_dict,
            collections.OrderedDict: self._encode_dict,
            list: self._encode_list,
            set: self._encode_set,
            tuple: self._encode_tuple,
            type(None): self._encode_none,
        }[o]



from common.configure_logging import redirect_logging_to_freeorion_logger

# Logging is redirected before other imports so that import errors appear in log files.
redirect_logging_to_freeorion_logger()

import freeorion as fo
import sys
from math import cos, pi, sin
from random import choice, random, uniform

from universe_tables import MONSTER_FREQUENCY


def execute_turn_events():
    print("Executing turn events for turn", fo.current_turn())

    # creating fields
    systems = fo.get_systems()
    radius = fo.get_universe_width() / 2.0
    field_types = ["FLD_MOLECULAR_CLOUD", "FLD_ION_STORM", "FLD_NANITE_SWARM", "FLD_METEOR_BLIZZARD", "FLD_VOID_RIFT"]

    if random() < max(0.00015 * radius, 0.03):
        field_type = choice(field_types)
        size = 5.0
        dist_from_center = uniform(0.35, 1.0) * radius
        angle = random() * 2.0 * pi
        x = radius + (dist_from_center * sin(angle))
        y = radius + (dist_from_center * cos(angle))

        print("...creating new", field_type, "field, at distance", dist_from_center, "from center")
        if fo.create_field(field_type, x, y, size) == fo.invalid_object():
            print("Turn events: couldn't create new field", file=sys.stderr)

    # creating monsters
    gsd = fo.get_galaxy_setup_data()
    monster_freq = MONSTER_FREQUENCY[gsd.monsterFrequency]
    # monster freq ranges from 0.0 .. 1/30 (= one monster per 30 systems) .. 1/3 (= one monster per 3 systems) .. 99/100 (almost everywhere)
    # (example: low monsters and 150 Systems results in 150 / 30 * 0.01 = 0.05)
    if monster_freq > 0 and random() < len(systems) * monster_freq * 0.01:
        # only spawn Krill at the moment, other monsters can follow in the future
        if random() < 1:
            monster_type = "SM_KRILL_1"
        else:
            monster_type = "SM_FLOATER"

        # search for systems without planets or fleets
        candidates = [s for s in systems if len(fo.sys_get_planets(s)) <= 0 and len(fo.sys_get_fleets(s)) <= 0]
        if not candidates:
            print("Turn events: unable to find system for monster spawn", file=sys.stderr)
        else:
            system = choice(candidates)
            print("...creating new", monster_type, "at", fo.get_name(system))

            # create monster fleet
            monster_fleet = fo.create_monster_fleet(system)
            # if fleet creation fails, report an error
            if monster_fleet == fo.invalid_object():
                print("Turn events: unable to create new monster fleet", file=sys.stderr)
            else:
                # create monster, if creation fails, report an error
                monster = fo.create_monster(monster_type, monster_fleet)
                if monster == fo.invalid_object():
                    print("Turn events: unable to create monster in fleet", file=sys.stderr)

    return True



from focs._effects import (
    HasTag,
    OwnedBy,
    Planet,
    Ship,
    Source,
    StatisticIf,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_PSIONICS",
    description="LRN_PSIONICS_DESC",
    short_description="POLICY_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=500 * TECH_COST_MULTIPLIER
    - (
        250
        * TECH_COST_MULTIPLIER
        * StatisticIf(
            float,
            # no Source condition here. Empire needs to have a ship or planet with a telepathic species, but it need not be the capital / source object
            condition=(Planet() | Ship) & OwnedBy(empire=Source.Owner) &
            # @content_tag{TELEPATHIC} Decreases research cost of this tech for empires that own any object with this tag
            HasTag(name="TELEPATHIC"),
        )
    ),
    researchturns=4,
    tags=["PEDIA_LEARNING_CATEGORY", "THEORY"],
    prerequisites=["LRN_TRANSLING_THT"],
    unlock=[
        Item(type=UnlockPolicy, name="PLC_INDOCTRINATION"),
        Item(type=UnlockPolicy, name="PLC_CONFORMANCE"),
    ],
    graphic="icons/tech/psionics.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_SMALL_ROBO",
    description="SHP_SMALL_ROBO_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=16 * TECH_COST_MULTIPLIER,
    researchturns=2,
    tags=["PEDIA_ROBOTIC_HULL_TECHS"],
    prerequisites=["SHP_MIL_ROBO_CONT", "PRO_NANOTECH_PROD"],
    unlock=Item(type=UnlockShipHull, name="SH_SMALL_ROBOTIC"),
    graphic="icons/ship_hulls/generic-medium-hull_small.png",
)



import freeorion as fo

import natives
import planets
import universe_tables
from common.print_utils import Number, Sequence, Table, Text

species_summary = {species: 0 for species in fo.get_native_species()}
empire_species = {species: 0 for species in fo.get_playable_species()}
potential_native_planet_summary = {planet_type: 0 for planet_type in planets.planet_types}
settled_native_planet_summary = {planet_type: 0 for planet_type in planets.planet_types}
monsters_summary = []
tracked_monsters_chance = {}
tracked_monsters_tries = {}
tracked_monsters_summary = {}
tracked_monsters_location_summary = {}
tracked_nest_location_summary = {}
specials_summary = {special: 0 for special in fo.get_all_specials()}
specials_repeat_dist = {count: 0 for count in [0, 1, 2, 3, 4]}


def log_planet_count_dist(sys_list):
    planet_count_dist = {}
    planet_size_dist = {size: 0 for size in planets.planet_sizes}
    for system in sys_list:
        planet_count = 0
        for planet in fo.sys_get_planets(system):
            this_size = fo.planet_get_size(planet)
            if this_size in planets.planet_sizes:
                planet_count += 1
                planet_size_dist[this_size] += 1
        planet_count_dist.setdefault(planet_count, [0])[0] += 1
    planet_tally = sum(planet_size_dist.values())

    count_distribution_table = Table(
        Text("planets in system"),
        Text("num of systems"),
        Number("% of systems", precession=1),
        table_name="Planet Count Distribution",
    )
    for planet_count, sys_count in planet_count_dist.items():
        count_distribution_table.add_row(
            planet_count,
            sys_count[0],
            100.0 * sys_count[0] / len(sys_list),
        )
    print(count_distribution_table)
    print()

    size_distribution = Table(
        Text("size"),
        Text("count"),
        Number("% of planets", precession=1),
        table_name="Planet Size Distribution",
    )
    for planet_size, planet_count in sorted(planet_size_dist.items()):
        size_distribution.add_row(
            planet_size,
            planet_count,
            100.0 * planet_count / planet_tally,
        )
    print(size_distribution)
    print()


def log_planet_type_summary(sys_list):
    planet_type_summary_table = {k: 0 for k in planets.planet_types}
    for system in sys_list:
        for planet in fo.sys_get_planets(system):
            planet_type_summary_table[fo.planet_get_type(planet)] += 1
    planet_total = sum(planet_type_summary_table.values())

    type_summary_table = Table(
        Text("planet type", align="<"),
        Number("% of planets", precession=1),
        table_name="Planet Type Summary for a total of %d placed planets" % planet_total,
    )

    for planet_type, planet_count in sorted(planet_type_summary_table.items()):
        type_summary_table.add_row(
            planet_type.name,
            100.0 * planet_count / planet_total,
        )
    print(type_summary_table)
    print()


def log_species_summary(native_freq):
    num_empires = sum(empire_species.values())
    num_species = len(fo.get_playable_species())
    exp_count = num_empires // num_species

    species_summary_table = Table(
        Text("species"),
        Text("count"),
        Number("%", precession=1),
        table_name=(
            "Empire Starting Species Summary\n"
            "Approximately %d to %d empires expected per species\n"
            "%d Empires and %d playable species"
        )
        % (max(0, exp_count - 1), exp_count + 1, num_empires, num_species),
    )

    for species, count in sorted(empire_species.items()):
        species_summary_table.add_row(
            species,
            count,
            100.0 * count / num_empires,
        )
    print(species_summary_table)
    print()

    native_chance = universe_tables.NATIVE_FREQUENCY[native_freq]
    # as the value in the universe table is higher for a lower frequency, we have to invert it
    # a value of 0 means no natives, in this case return immediately
    if native_chance <= 0:
        return
    native_table = Table(
        Text("settled natives"),
        Text("on planets"),
        Number("total", precession=1),
        Number("actual", precession=1),
        Number("expected", precession=1),
        Sequence("planet types"),
        table_name="Natives Placement Summary (native frequency: %.1f%%)" % (100 * native_chance),
    )

    native_potential_planet_total = sum(potential_native_planet_summary.values())
    for species in sorted(species_summary):
        if species_summary[species] > 0:
            settleable_planets = 0
            expectation_tally = 0.0
            for p_type in natives.planet_types_for_natives[species]:
                settleable_planets += potential_native_planet_summary[p_type]
                expectation_tally += (
                    native_chance
                    * 100.0
                    * potential_native_planet_summary[p_type]
                    / (1e-10 + len(natives.natives_for_planet_type[p_type]))
                )
            expectation = expectation_tally / (1e-10 + settleable_planets)
            native_table.add_row(
                species,
                species_summary[species],
                100.0 * species_summary[species] / (1e-10 + native_potential_planet_total),
                100.0 * species_summary[species] / (1e-10 + settleable_planets),
                expectation,
                [str(p_t) for p_t in natives.planet_types_for_natives[species]],
            )

    print(native_table)
    print()

    native_settled_planet_total = sum(settled_native_planet_summary.values())
    type_summary_table = Table(
        Text("planet type"),
        Number("potential (% of tot)", precession=1),
        Number("settled (% of potential)", precession=1),
        table_name=(
            "Planet Type Summary for Native Planets (native frequency: {:.1f}%)\n"
            "Totals: native_potential_planet_total: {}; native_settled_planet_total {}"
        ).format(100 * native_chance, native_potential_planet_total, native_settled_planet_total),
    )

    for planet_type, planet_count in sorted(potential_native_planet_summary.items()):
        settled_planet_count = settled_native_planet_summary.get(planet_type, 0)
        potential_percent = 100.0 * planet_count / (1e-10 + native_potential_planet_total)
        settled_percent = 100.0 * settled_planet_count / (1e-10 + planet_count)
        type_summary_table.add_row(
            planet_type.name,
            potential_percent,
            settled_percent,
        )
    print(type_summary_table)
    print()


def log_monsters_summary(monster_freq):
    monster_place_table = Table(
        Text("monster"),
        Text("count"),
        table_name="Monster placement",
    )
    for monster, counter in sorted(monsters_summary):
        monster_place_table.add_row(monster, counter)
    print(monster_place_table)
    print()

    monster_chance = universe_tables.MONSTER_FREQUENCY[monster_freq]
    monster_table = Table(
        Text("monster"),
        Number("chance"),
        Text("attempts"),
        Text("number placed"),
        Text("number valid sys locs"),
        Text("number valid nest planet locs"),
        table_name=(
            "Space Monsters Placement Summary\n"
            "Tracked Monster and Nest Summary (base monster freq: %4.1f%%)" % (100 * monster_chance)
        ),
    )
    for monster in sorted(tracked_monsters_summary):
        monster_table.add_row(
            monster,
            tracked_monsters_chance[monster],
            tracked_monsters_tries[monster],
            tracked_monsters_summary[monster],
            tracked_monsters_location_summary[monster],
            tracked_nest_location_summary[monster],
        )
    print(monster_table)
    print()


def log_specials_summary():
    special_placement_count_table = Table(
        Text("special"),
        Text("times placed"),
        table_name="Special Placement Summary",
    )
    for special in sorted(specials_summary):
        special_placement_count_table.add_row(special, specials_summary[special])
    print(special_placement_count_table)
    print()

    special_placement = Table(
        Text("count"),
        Text("tally"),
        Number("% of objects", precession=1),
        table_name="Specials Count(Repeat) Distribution",
    )
    objects_tally = sum(specials_repeat_dist.values())
    for number, tally in specials_repeat_dist.items():
        special_placement.add_row(
            number,
            tally,
            100.0 * tally / (1e-10 + objects_tally),
        )
    print(special_placement)
    print()


def log_systems():
    universe = fo.get_universe()

    systems_table = Table(
        Text("id"),
        Text("name"),
        Sequence("planets"),
        Sequence("connections"),
        Text("star"),
        table_name="System summary",
    )
    for sid in fo.get_systems():
        system = universe.getSystem(sid)
        systems_table.add_row(
            sid,
            system.name,
            fo.sys_get_planets(sid),
            fo.sys_get_starlanes(sid),
            system.starType.name,
        )
    print(systems_table)


def log_planets():
    universe = fo.get_universe()
    planets_table = Table(
        Text("id"),
        Text("name"),
        Text("system"),
        Text("type"),
        Sequence("specials"),
        Text("species"),
        Sequence("buildings"),
        table_name="Planets summary",
    )
    # group planets by system
    for sid in fo.get_systems():
        for pid in fo.sys_get_planets(sid):
            planet = universe.getPlanet(pid)

            planet_type = fo.planet_get_type(pid).name
            planet_size = fo.planet_get_size(pid).name
            if planet_type != planet_size:
                planet_type = f"{planet_type} {planet_size}"

            buildings = [universe.getBuilding(x).name for x in planet.buildingIDs]
            planets_table.add_row(
                pid,
                planet.name,
                planet.systemID,
                planet_type,
                list(planet.specials),
                planet.speciesName,
                buildings,
            )

    print(planets_table)



from inspect import getdoc, isroutine
from logging import error, warning


def _get_member_info(name, member):
    type_ = str(type(member))

    info = {
        "type": type_,
    }
    if isinstance(member, property):
        info["getter"] = getdoc(member.fget)
    elif isroutine(member):
        info["routine"] = (member.__name__, getdoc(member))
    elif isinstance(member, (int, str, float, list, tuple, dict, set, frozenset)):
        pass  # we don't need any
    elif "freeOrionAIInterface" in type_ or "freeorion" in type_:
        pass  # TODO we got some instance here, probably we should inspect it too.
    else:
        # instance properties will be already resolved
        warning('[%s] Unexpected member "%s"(%s)', name, type(member), member)
    return info


def _getmembers(obj, predicate=None):
    """Return all members of an object as (name, value) pairs sorted by name.
    Optionally, only return members that satisfy a given predicate."""
    results = []
    for key in dir(obj):
        try:
            value = getattr(obj, key)
        except AttributeError:
            continue
        except Exception as e:
            message = [
                "-" * 20,
                f'Error in "{obj.__class__.__name__}.{key}" with error',
                "..." * 20,
                "Error info:",
                str(e),
                "..." * 20,
            ]
            error("\n".join(message))
            continue
        if not predicate or predicate(value):
            results.append((key, value))
    results.sort()
    return results



from focs._effects import (
    CanColonize,
    EffectsGroup,
    GasGiantType,
    GenerateSitRepMessage,
    HasSpecies,
    Huge,
    IsSource,
    Large,
    LocalCandidate,
    MinOf,
    OwnedBy,
    Planet,
    SetMaxSupply,
    SetSupply,
    Small,
    Source,
    Target,
    Tiny,
    Turn,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.misc import GROWTH_RATE_FACTOR, MIN_RECOLONIZING_HAPPINESS, MIN_RECOLONIZING_SIZE
from macros.priorities import AFTER_ALL_TARGET_MAX_METERS_PRIORITY

Tech(
    name="SHP_GAL_EXPLO",
    description="SHP_GAL_EXPLO_DESC",
    short_description="POLICY_UNLOCK_SHORT_DESC",
    category="SHIP_PARTS_CATEGORY",
    researchcost=1 * TECH_COST_MULTIPLIER,
    researchturns=1,
    tags=["PEDIA_SHIP_PARTS_CATEGORY", "THEORY"],
    unlock=Item(type=UnlockPolicy, name="PLC_EXPLORATION"),
    effectsgroups=[
        EffectsGroup(
            scope=Planet(size=[Tiny]) & OwnedBy(empire=Source.Owner),
            accountinglabel="TINY_PLANET_LABEL",
            effects=SetMaxSupply(value=Value + 2),
        ),
        EffectsGroup(
            scope=Planet(size=[Small]) & OwnedBy(empire=Source.Owner),
            accountinglabel="SMALL_PLANET_LABEL",
            effects=SetMaxSupply(value=Value + 1),
        ),
        EffectsGroup(
            scope=Planet(size=[Large]) & OwnedBy(empire=Source.Owner),
            accountinglabel="LARGE_PLANET_LABEL",
            effects=SetMaxSupply(value=Value - 1),
        ),
        EffectsGroup(
            scope=Planet(size=[Huge]) & OwnedBy(empire=Source.Owner),
            accountinglabel="HUGE_PLANET_LABEL",
            effects=SetMaxSupply(value=Value - 2),
        ),
        EffectsGroup(
            scope=Planet(type=[GasGiantType]) & OwnedBy(empire=Source.Owner),
            accountinglabel="GAS_GIANT_LABEL",
            effects=SetMaxSupply(value=Value - 1),
        ),
        EffectsGroup(  # outpost supply increases 1 per turn up to max
            scope=Planet() & OwnedBy(empire=Source.Owner) & ~HasSpecies(),
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetSupply(value=MinOf(float, Value(Target.MaxSupply), Value + 1)),
        ),
        # generate sitrep for any planet that is about to increase to above the
        # recolonizing population, while also having sufficient happiness to do
        # do so
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & CanColonize
            & (LocalCandidate.Population < LocalCandidate.TargetPopulation)
            & (
                LocalCandidate.Population
                + LocalCandidate.Population
                * GROWTH_RATE_FACTOR
                * (LocalCandidate.TargetPopulation + 1 - LocalCandidate.Population)
                >= MIN_RECOLONIZING_SIZE
            )
            & (LocalCandidate.TargetHappiness >= MIN_RECOLONIZING_HAPPINESS)
            & (LocalCandidate.Happiness >= MIN_RECOLONIZING_HAPPINESS - 1)
            & (
                (0.1 <= LocalCandidate.Population) & (LocalCandidate.Population <= (MIN_RECOLONIZING_SIZE - 0.001))
                | (0.1 <= LocalCandidate.Happiness) & (LocalCandidate.Happiness <= (MIN_RECOLONIZING_HAPPINESS - 0.001))
            ),
            effects=[
                GenerateSitRepMessage(
                    message="SITREP_POP_THRESHOLD",
                    label="SITREP_POP_THRESHOLD_LABEL",
                    icon="icons/sitrep/colony_growth.png",
                    parameters={"planet": Target.ID},
                    empire=Source.Owner,
                )
            ],
        ),
        EffectsGroup(
            scope=IsSource,
            activation=Turn(low=0, high=0),
            effects=[
                GenerateSitRepMessage(
                    message="SITREP_WELCOME",
                    label="SITREP_WELCOME_LABEL",  # explicitly provided so that the custom SitRep intro message can share the same SitRepPanel filter
                    icon="icons/sitrep/fo_logo.png",
                    empire=Source.Owner,
                )
            ],
        ),
    ],
    graphic="icons/tech/galactic_exploration.png",
)



from focs._effects import (
    Capital,
    Contains,
    EmpireHasAdoptedPolicy,
    HasSpecial,
    HasSpecies,
    HasTag,
    Homeworld,
    IsBuilding,
    LocalCandidate,
    Mode,
    OwnedBy,
    OwnerHasTech,
    Planet,
    Source,
    Statistic,
)
from focs._species import *

HAS_INDUSTRY_FOCUS = FocusType(
    name="FOCUS_INDUSTRY", description="FOCUS_INDUSTRY_DESC", location=Planet(), graphic="icons/focus/industry.png"
)

HAS_RESEARCH_FOCUS = FocusType(
    name="FOCUS_RESEARCH",
    description="FOCUS_RESEARCH_DESC",
    location=Planet()
    & (
        ~EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_RACIAL_PURITY")
        | HasSpecies(name=["SP_EXOBOT"])
        | HasSpecies(
            name=[Statistic(str, Mode, value=LocalCandidate.Species, condition=Capital & OwnedBy(empire=Source.Owner))]
        )
    ),
    graphic="icons/focus/research.png",
)

HAS_INFLUENCE_FOCUS = FocusType(
    name="FOCUS_INFLUENCE",
    description="FOCUS_INFLUENCE_DESC",
    location=Planet()
    & (
        ~EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_RACIAL_PURITY")
        | HasSpecies(name=["SP_EXOBOT"])
        | HasSpecies(
            name=[Statistic(str, Mode, value=LocalCandidate.Species, condition=Capital & OwnedBy(empire=Source.Owner))]
        )
    ),
    graphic="icons/focus/influence.png",
)

HAS_GROWTH_FOCUS = FocusType(
    name="FOCUS_GROWTH",
    description="FOCUS_GROWTH_DESC",
    location=(Homeworld(name=[Source.Species]) & ~HasTag(name="SELF_SUSTAINING"))
    | HasSpecial(name="POSITRONIUM_SPECIAL")
    | HasSpecial(name="SUPERCONDUCTOR_SPECIAL")
    | HasSpecial(name="MONOPOLE_SPECIAL")
    | HasSpecial(name="SPICE_SPECIAL")
    | HasSpecial(name="FRUIT_SPECIAL")
    | HasSpecial(name="PROBIOTIC_SPECIAL")
    | HasSpecial(name="ELERIUM_SPECIAL")
    | HasSpecial(name="CRYSTALS_SPECIAL")
    | HasSpecial(name="MINERALS_SPECIAL"),
    graphic="icons/focus/growth.png",
)

HAS_ADVANCED_FOCI = [
    FocusType(
        name="FOCUS_PROTECTION",
        description="FOCUS_PROTECTION_DESC",
        # location = OwnerHasTech name = "DEF_ROOT_DEFENSE"
        location=Planet(),  # empires have self-defense anyway, and this way some natives can now defend themselves better
        graphic="icons/focus/protection.png",
    ),
    FocusType(
        name="FOCUS_LOGISTICS",
        description="FOCUS_LOGISTICS_DESC",
        location=OwnerHasTech(name="SHP_INTSTEL_LOG"),
        graphic="icons/focus/supply.png",
    ),
    FocusType(
        name="FOCUS_STOCKPILE",
        description="FOCUS_STOCKPILE_DESC",
        location=OwnerHasTech(name="PRO_GENERIC_SUPPLIES"),
        graphic="icons/focus/stockpile.png",
    ),
    FocusType(
        name="FOCUS_STEALTH",
        description="FOCUS_STEALTH_DESC",
        location=Contains(IsBuilding(name=["BLD_PLANET_CLOAK"]))
        | (Contains(IsBuilding(name=["BLD_TRANSFORMER"])) & OwnerHasTech(name="DEF_PLANET_CLOAK")),
        graphic="icons/focus/stealth.png",
    ),
    FocusType(
        name="FOCUS_BIOTERROR",
        description="FOCUS_BIOTERROR_DESC",
        location=Contains(IsBuilding(name=["BLD_BIOTERROR_PROJECTOR"]))
        | (Contains(IsBuilding(name=["BLD_TRANSFORMER"])) & OwnerHasTech(name="GRO_BIOTERROR")),
        graphic="icons/focus/bioterror.png",
    ),
    FocusType(
        name="FOCUS_STARGATE_SEND",
        description="FOCUS_STARGATE_SEND_DESC",
        location=Contains(IsBuilding(name=["BLD_STARGATE"]))
        | (Contains(IsBuilding(name=["BLD_TRANSFORMER"])) & OwnerHasTech(name="CON_STARGATE")),
        graphic="icons/focus/stargate_send.png",
    ),
    FocusType(
        name="FOCUS_STARGATE_RECEIVE",
        description="FOCUS_STARGATE_RECEIVE_DESC",
        location=Contains(IsBuilding(name=["BLD_STARGATE"]))
        | (Contains(IsBuilding(name=["BLD_TRANSFORMER"])) & OwnerHasTech(name="CON_STARGATE")),
        graphic="icons/focus/stargate_receive.png",
    ),
    FocusType(
        name="FOCUS_PLANET_DRIVE",
        description="FOCUS_PLANET_DRIVE_DESC",
        location=Contains(IsBuilding(name=["BLD_PLANET_DRIVE"]))
        | (Contains(IsBuilding(name=["BLD_TRANSFORMER"])) & OwnerHasTech(name="CON_PLANET_DRIVE")),
        graphic="icons/building/planetary_stardrive.png",
    ),
    FocusType(
        name="FOCUS_DISTORTION",
        description="FOCUS_DISTORTION_DESC",
        location=Contains(IsBuilding(name=["BLD_SPATIAL_DISTORT_GEN"]))
        | (Contains(IsBuilding(name=["BLD_TRANSFORMER"])) & OwnerHasTech(name="LRN_SPATIAL_DISTORT_GEN")),
        graphic="icons/focus/distortion.png",
    ),
    FocusType(
        name="FOCUS_DOMINATION",
        description="FOCUS_DOMINATION_DESC",
        location=HasTag(name="TELEPATHIC") & OwnerHasTech(name="LRN_PSY_DOM"),
        graphic="icons/focus/psi_domination.png",
    ),
]



from abc import ABCMeta, abstractmethod


class ChatHandlerBase(metaclass=ABCMeta):
    """
    Base class for chat handling.
    """

    def __init__(self):
        self.debug_active = False  # switched to true when any AI is debugged to stop normal chat reactions

    @abstractmethod
    def process_message(self, sender_id: int, message: str) -> bool:
        """
        Process chat message and return if message was processed, and should not be send to another handlers.
        """



from focs._effects import EffectsGroup, IsBuilding, OwnedBy, SetStealth, Source, Value
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_FRC_ENRG_CAMO",
    description="CON_FRC_ENRG_CAMO_DESC",
    short_description="STEALTH_SHORT_DESC",
    category="SPY_CATEGORY",
    researchcost=180 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_SPY_CATEGORY"],
    prerequisites=["CON_FRC_ENRG_STRC"],
    effectsgroups=[
        EffectsGroup(scope=IsBuilding() & OwnedBy(empire=Source.Owner), effects=SetStealth(value=Value + 20))
    ],
)



from abc import ABC

from common.statistic_interface._serizlizer import (
    DictSerializer,
    Serializable,
    Serializer,
    to_float,
    to_int,
    to_str,
)


class Dump(ABC):
    def __init__(self, name: str, serializer: Serializer, multi=False):
        self.name = name
        self.multi = multi
        self._serializer = serializer

    def deserialize(self, value: str) -> Serializable:
        return self._serializer.deserialize(value)

    def serialize(self, value: Serializable) -> str:
        return self._serializer.serialize(value)


class DumpInt(Dump):
    def __init__(self, name: str):
        super().__init__(name, to_int)


class DumpFloat(Dump):
    def __init__(self, name: str):
        super().__init__(name, to_float)


class DumpStr(Dump):
    def __init__(self, name: str):
        super().__init__(name, to_str)


class DumpDict(Dump):
    def __init__(self, name: str, items: dict[str, Serializer]):
        super().__init__(name, DictSerializer(items))


class DumpMultiple(Dump):
    def __init__(self, name: str, serializer: Serializer):
        super().__init__(name, serializer, multi=True)



import pytest

from freeorion_tools.lazy_initializer import InitializerLock


@pytest.fixture
def lock():
    lock = InitializerLock("test")
    return lock


def test_function_return_value_when_unlocked(lock):
    @lock
    def foo():
        return 1

    lock.unlock()
    assert foo() == 1


def test_function_call_raise_error_when_locked(lock):
    @lock
    def foo():
        return 1

    with pytest.raises(ValueError, match=r"Call of 'foo' is forbidden before InitializerLock\('test'\) is initialized"):
        foo()



class EnumInfo:
    def __init__(self, name: str, attributes: dict[str, int]):
        self.name = name
        self.attributes = attributes


def inspect_enum(name, obj):
    return EnumInfo(name, {k: v.numerator for k, v in obj.names.items()})



import custom_sitreps
from focs._tech import *

Tech(
    name="SPY_CUSTOM_ADVISORIES",
    description="SPY_CUSTOM_ADVISORIES_DESC",
    short_description="SPY_CUSTOM_ADVISORIES",
    category="SPY_CATEGORY",
    researchcost=1,
    researchturns=1,
    tags=["PEDIA_SPY_CATEGORY"],
    effectsgroups=custom_sitreps.effectsgroups,
)



import freeOrionAIInterface as fo
import math
import random
from collections.abc import Iterable, Iterator
from typing import Optional, Union

import ShipDesignAI
from empire.ship_builders import get_shipyards
from EnumsAI import PriorityType
from freeorion_tools.caching import cache_for_current_turn
from freeorion_tools.timers import AITimer
from turn_state._planet_state import get_inhabited_planets


@cache_for_current_turn
def get_design_repository() -> dict[PriorityType, tuple[float, int, int, float]]:
    """Calculate the best designs for each ship class available at this turn."""
    design_repository = {}  # dict of tuples (rating,pid,designID,cost) sorted by rating and indexed by priority type

    design_timer = AITimer("ShipDesigner")
    design_timer.start("Updating cache for new turn")

    # TODO Don't use PriorityType but introduce more reasonable Enum
    designers = [
        ("Orbital Invasion", PriorityType.PRODUCTION_ORBITAL_INVASION, ShipDesignAI.OrbitalTroopShipDesigner),
        ("Invasion", PriorityType.PRODUCTION_INVASION, ShipDesignAI.StandardTroopShipDesigner),
        (
            "Orbital Colonization",
            PriorityType.PRODUCTION_ORBITAL_COLONISATION,
            ShipDesignAI.OrbitalColonisationShipDesigner,
        ),
        ("Colonization", PriorityType.PRODUCTION_COLONISATION, ShipDesignAI.StandardColonisationShipDesigner),
        ("Orbital Outposter", PriorityType.PRODUCTION_ORBITAL_OUTPOST, ShipDesignAI.OrbitalOutpostShipDesigner),
        ("Outposter", PriorityType.PRODUCTION_OUTPOST, ShipDesignAI.StandardOutpostShipDesigner),
        ("Orbital Defense", PriorityType.PRODUCTION_ORBITAL_DEFENSE, ShipDesignAI.OrbitalDefenseShipDesigner),
        ("Scouts", PriorityType.PRODUCTION_EXPLORATION, ShipDesignAI.ScoutShipDesigner),
    ]

    for timer_name, priority_type, designer in designers:
        design_timer.start(timer_name)
        design_repository[priority_type] = designer().optimize_design()
    best_military_stats = ShipDesignAI.WarShipDesigner().optimize_design()
    best_carrier_stats = ShipDesignAI.CarrierShipDesigner().optimize_design()
    best_stats = best_military_stats + best_carrier_stats if random.random() < 0.8 else best_military_stats
    best_stats.sort(reverse=True)
    design_repository[PriorityType.PRODUCTION_MILITARY] = best_stats
    design_timer.start("Krill Spawner")
    ShipDesignAI.KrillSpawnerShipDesigner().optimize_design()  # just designing it, building+mission not supported yet
    if fo.currentTurn() % 10 == 0:
        design_timer.start("Printing")
        ShipDesignAI.Cache.print_best_designs()
    design_timer.stop_print_and_clear()
    return design_repository


@cache_for_current_turn
def cur_best_military_design_rating() -> float:
    """
    Find and return the default combat rating of our best military design.
    """
    design_repository = get_design_repository()

    priority = PriorityType.PRODUCTION_MILITARY
    if design_repository.get(priority, None) and design_repository[priority][0]:
        # the rating provided by the ShipDesigner does not
        # reflect the rating used in threat considerations
        # but takes additional factors (such as cost) into
        # account. Therefore, we want to calculate the actual
        # rating of the design as provided by CombatRatingsAI.
        _, _, _, _, stats = design_repository[priority][0]
        # TODO: Should this consider enemy stats?
        rating = stats.convert_to_combat_stats().get_rating()
        return max(rating, 0.001)
    return 0.001


def _get_locations(locations: Union[int, Iterable[int], None]) -> frozenset[int]:
    if locations is None:
        return get_inhabited_planets()

    if isinstance(locations, int):
        locations = (locations,)
    return get_inhabited_planets().intersection(locations)


def get_best_ship_info(
    priority: PriorityType, loc: Union[int, Iterable[int], None] = None
) -> tuple[Optional[int], Optional["fo.shipDesign"], Optional[list[int]]]:
    """Returns 3 item tuple: designID, design, buildLocList."""
    planet_ids = _get_locations(loc)
    if not planet_ids:
        return None, None, None
    return _get_best_ship_info(priority, tuple(planet_ids))


def _get_best_ship_info(
    priority: PriorityType, planet_ids: tuple[int]
) -> tuple[Optional[int], Optional["fo.shipDesign"], Optional[list[int]]]:
    design_repository = get_design_repository()

    if priority in design_repository:
        best_designs = design_repository[priority]
        if not best_designs:
            return None, None, None

        # best_designs are already sorted by rating high to low, so the top rating is the first encountered within
        # our planet search list
        for design_stats in best_designs:
            top_rating, pid, top_id, cost, stats = design_stats
            if pid in planet_ids:
                break
        else:
            return None, None, None  # apparently can't build for this priority within the desired planet group
        valid_locs = [
            pid_
            for rating, pid_, design_id, _, _ in best_designs
            if rating == top_rating and design_id == top_id and pid_ in planet_ids
        ]
        return top_id, fo.getShipDesign(top_id), valid_locs
    else:
        return None, None, None  # must be missing a Shipyard or other orbital (or missing tech)


def get_best_ship_ratings(planet_ids: tuple[int]) -> list[tuple[float, int, int, "fo.shipDesign"]]:
    """
    Returns list of [partition, pid, designID, design] tuples, currently only for military ships.

    Since we haven't yet implemented a way to target military ship construction at/near particular locations
    where they are most in need, and also because our rating system is presumably useful-but-not-perfect, we want to
    distribute the construction across the Resource Group and across similarly rated designs, preferentially choosing
    the best rated design/loc combo, but if there are multiple design/loc combos with the same or similar ratings then
    we want some chance of choosing  those alternate designs/locations.

    The approach to this taken below is to treat the ratings akin to an energy to be used in a statistical mechanics
    type partition function. 'tally' will compute the normalization constant.
    So first go through and calculate the tally as well as convert each individual contribution to
    the running total up to that point, to facilitate later sampling.  Then those running totals are
    renormalized by the final tally, so that a later random number selector in the range [0,1) can be
    used to select the chosen design/loc.
    """
    return list(_get_best_ship_ratings(tuple(planet_ids)))


@cache_for_current_turn
def _get_best_ship_ratings(planet_ids: tuple[int]) -> Iterator[tuple[float, int, int, "fo.shipDesign"]]:
    design_repository = get_design_repository()
    priority = PriorityType.PRODUCTION_MILITARY
    planet_ids = set(planet_ids).intersection(get_shipyards())

    if priority not in design_repository:
        return

    build_choices = design_repository[priority]
    loc_choices = [
        [rating, pid, design_id, fo.getShipDesign(design_id)]
        for (rating, pid, design_id, cost, stats) in build_choices
        if pid in planet_ids
    ]
    if not loc_choices:
        return
    best_rating = loc_choices[0][0]
    tally = 0
    ret_val = []
    for rating, pid, design_id, design in loc_choices:
        if rating < 0.7 * best_rating:
            break
        p = math.exp(10 * (rating / best_rating - 1))
        tally += p
        ret_val.append((tally, pid, design_id, design))
    for base_tally, pid, design_id, design in ret_val:
        yield base_tally / tally, pid, design_id, design
